r"""
Simple custom PDE class
=======================

This example implements a custom scalar PDE. To display some of the features of
the package, we here implemented the `Kuramoto–Sivashinsky equation
<https://en.wikipedia.org/wiki/Kuramoto–Sivashinsky_equation>`_, which was
considered for flame front dynamics:

.. math::
    \partial_t u = -\frac12 |\nabla u|^2 - \nabla^2 u - \nabla^4 u
"""

from pde import UnitGrid, ScalarField, PDEBase


class KuramotoSivashinskyPDE(PDEBase):
    """ Implementation of the normalized Kuramoto–Sivashinsky equation """
    
    def evolution_rate(self, state, t=0):
        """ implement the python version of the evolution equation """
        state_lap = state.laplace(bc='natural')
        state_lap2 = state_lap.laplace(bc='natural')
        state_grad = state.gradient(bc='natural')
        return -state_grad.to_scalar('squared_sum') / 2 - state_lap - state_lap2 


grid = UnitGrid([16, 16])  # generate grid
state = ScalarField.random_uniform(grid)  # generate initial condition

eq = KuramotoSivashinskyPDE()  # define the pde
result = eq.solve(state, t_range=10, dt=0.01)
result.plot(show=True)



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np
import pytest

from ...fields import ScalarField
from ...grids import UnitGrid
from ...pdes import DiffusionPDE
from .. import Controller, ExplicitSolver, ImplicitSolver, ScipySolver
from .. import registered_solvers



def test_solver_registration():
    """ test solver registration """
    solvers = registered_solvers()
    assert 'explicit' in solvers
    assert 'scipy' in solvers



def test_solver_in_pde_class():
    """ test whether solver instances can be used in pde instances """
    field = ScalarField.random_uniform(UnitGrid([16, 16]), -1, 1)
    eq = DiffusionPDE()
    eq.solve(field, t_range=1, method=ScipySolver)



@pytest.mark.parametrize('solver_class', [ExplicitSolver, ImplicitSolver,
                                          ScipySolver])
def test_compare_solvers(solver_class):
    """ compare several solvers """
    field = ScalarField.random_uniform(UnitGrid([8, 8]), -1, 1)
    eq = DiffusionPDE()
    
    # ground truth
    c1 = Controller(ExplicitSolver(eq, scheme='runge-kutta'), t_range=.1,
                    tracker=None)
    s1 = c1.run(field, dt=5e-3)
    
    c2 = Controller(solver_class(eq), t_range=.1, tracker=None)
    with np.errstate(under='ignore'):
        s2 = c2.run(field, dt=5e-3)
                           
    np.testing.assert_allclose(s1.data, s2.data, rtol=1e-2, atol=1e-2)




'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import glob
import sys
import os
import subprocess as sp
from pathlib import Path
from typing import List  # @UnusedImport

import pytest
import numba as nb

from ..tools.misc import module_available, skipUnlessModule



PACKAGE_PATH = Path(__file__).resolve().parents[2]
EXAMPLES = glob.glob(str(PACKAGE_PATH / 'examples' / '*.py'))
NOTEBOOKS = glob.glob(str(PACKAGE_PATH / 'examples' / 'jupyter' / '*.ipynb'))

SKIP_EXAMPLES: List[str] = ['make_movie.py']
if not module_available("matplotlib"):
    SKIP_EXAMPLES.append('trackers.py')


@pytest.mark.skipif(sys.platform == 'win32', reason="Assumes unix setup")
@pytest.mark.skipif(nb.config.DISABLE_JIT,
                    reason='pytest seems to check code coverage')
@pytest.mark.parametrize('path', EXAMPLES)
def test_example(path):
    """ runs an example script given by path """
    if os.path.basename(path).startswith('_'):
        pytest.skip('skip examples starting with an underscore')
    if any(name in path for name in SKIP_EXAMPLES):
        pytest.skip(f'Skip test {path}')
    
    env = os.environ.copy()
    env["PYTHONPATH"] = str(PACKAGE_PATH) + ":" + env.get("PYTHONPATH", "")
    proc = sp.Popen([sys.executable, path], env=env, stdout=sp.PIPE,
                    stderr=sp.PIPE)
    try:
        outs, errs = proc.communicate(timeout=30)
    except sp.TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()            
        
    msg = 'Script `%s` failed with following output:' % path
    if outs:
        msg = '%s\nSTDOUT:\n%s' % (msg, outs)
    if errs:
        msg = '%s\nSTDERR:\n%s' % (msg, errs)
    assert proc.returncode == 0, msg



@skipUnlessModule('jupyter')
@pytest.mark.skipif(nb.config.DISABLE_JIT,
                    reason='pytest seems to check code coverage')
@pytest.mark.parametrize('path', NOTEBOOKS)
def test_jupyter_notebooks(path, tmp_path):
    """ run the jupyter notebooks """
    if os.path.basename(path).startswith('_'):
        pytest.skip('skip examples starting with an underscore')
        
    # adjust python environment
    my_env = os.environ.copy()
    my_env["PYTHONPATH"] = str(PACKAGE_PATH) + ":" + my_env["PATH"]        
        
    outfile = tmp_path / os.path.basename(path)
    sp.check_call([sys.executable, '-m', 'jupyter', 'nbconvert', 
                   '--to', 'notebook', '--output', outfile,
                   '--execute', path], env=my_env)



'''
Functions for creating movies of simulation results


.. autosummary::
   :nosignatures:

   movie_scalar
   movie_multiple
   
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import functools
import os
import subprocess as sp
import tempfile
import shutil
from typing import Dict, Any

from .plotting import ScalarFieldPlot, ScaleData
from ..storage.base import StorageBase
from ..tools.docstrings import fill_in_docstring



class Movie:
    """ Class for creating movies from matplotlib figures using ffmpeg """

    def __init__(self, width=None, filename=None, verbose=False,
                 framerate=None, image_folder=None):
        self.width = width          # pixel width of the movie
        self.filename = filename    # filename used to save the movie
        self.verbose = verbose      # verbose encoding information?
        self.framerate = framerate  # framerate of the movie
        self.image_folder = image_folder  # folder where images are stored

        # internal variables
        self.recording = False
        self.frame = 0
        self._delete_images = False
        self._start()


    def __del__(self):
        self._end()


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_value, exc_tb):
        if self.filename is not None:
            self.save(self.filename)
        self._end()
        return False


    def _start(self):
        """ initializes the video recording """
        # create temporary directory for the image files of the movie
        if self.image_folder is None:
            self.image_folder = tempfile.mkdtemp(prefix='movie_')
            self._delete_images = True
        self.frame = 0
        self.recording = True


    def _end(self):
        """ clear up temporary things if necessary """
        if self.recording:
            if self._delete_images:
                shutil.rmtree(self.image_folder)
            self.recording = False


    def clear(self):
        """ delete current status and start from scratch """
        self._end()
        self._start()


    def _add_file(self, save_function):
        """
        Adds a file to the current movie
        """
        if not self.recording:
            raise ValueError('Movie is not initialized.')

        save_function("%s/frame_%09d.png" % (self.image_folder, self.frame))
        self.frame += 1


    def add_image(self, image):
        """
        Adds the data of a PIL image as a frame to the current movie.
        """
        self._add_file(image.save)


    def add_figure(self, fig=None):
        """ adds the figure `fig` as a frame to the current movie """
        import matplotlib.pyplot as plt
        if fig is None:
            fig = plt.gcf()

        if self.width is None:
            dpi = None
        else:
            dpi = self.width / fig.get_figwidth()

        # save image
        save_function = functools.partial(fig.savefig, dpi=dpi)
        self._add_file(save_function)


    def save_frames(self, filename_pattern='./frame_%09d.png', frames='all'):
        """ saves the given `frames` as images using the `filename_pattern` """
        if not self.recording:
            raise ValueError('Movie is not initialized.')

        if 'all' == frames:
            frames = range(self.frame)

        for f in frames:
            shutil.copy(
                "%s/frame_%09d.png" % (self.image_folder, f),
                str(filename_pattern) % f
            )


    def save(self, filename=None, extra_args=None):
        """ convert the recorded images to a movie using ffmpeg """
        if filename is None:
            if self.filename is None:
                raise ValueError('`filename` has to be supplied')
            filename = self.filename
        filename = os.path.expanduser(filename)
        
        if not self.recording:
            raise ValueError('Movie is not initialized.')

        # set parameters
        if extra_args is None:
            extra_args = []
        if self.framerate is not None:
            extra_args += ["-r", self.framerate]

        # construct the call to ffmpeg
        # add the `-pix_fmt yuv420p` switch for compatibility reasons
        #     -> http://ffmpeg.org/trac/ffmpeg/wiki/x264EncodingGuide
        args = ["ffmpeg"]
        if extra_args:
            args += extra_args
        args += [
            "-y",  # don't ask questions
            "-f", "image2",  # input format
            "-i", "%s/frame_%%09d.png" % self.image_folder,  # input data
            "-pix_fmt", "yuv420p",  # pixel format for compatibility
            "-b:v", "1024k",  # high bit rate for good quality
            str(filename)  # output file
        ]

        # spawn the subprocess and capture its output
        proc = sp.Popen(args, stdout=sp.PIPE, stderr=sp.PIPE)
        out, err = proc.communicate()

        # do output anyway, when verbosity is requested
        if self.verbose:
            print(out)
            print(err)

        return out, err



@fill_in_docstring
def movie_scalar(storage: StorageBase,
                 filename: str,
                 scale: ScaleData = 'automatic',
                 extras: Dict[str, Any] = None,
                 progress: bool = True,
                 tight: bool = False,
                 show: bool = True) -> None:
    """ produce a movie for a simulation of a scalar field
    
    Args:
        storage (:class:`~pde.storage.base.StorageBase`):
            The storage instance that contains all the data for the movie
        filename (str):
            The filename to which the movie is written. The extension determines
            the format used.
        scale (str, float, tuple of float):
            {ARG_PLOT_SCALE}
        extras (dict, optional):
            Additional functions that are evaluated and shown for each time 
            step. The key of the dictionary is used as a panel title.
        progress (bool):
            Flag determining whether the progress of making the movie is shown.
        tight (bool):
            Whether to call :func:`matplotlib.pyplot.tight_layout`. This affects
            the layout of all plot elements.
        show (bool):
            Flag determining whether images are shown during making the movie
    """
    quantities = [{'title': 'Concentration', 'source': None}]
    if extras:
        for key, value in extras.items():
            quantities.append({'title': key, 'source': value})
    
    # initialize the plot with the first data point
    plot = ScalarFieldPlot.from_storage(storage, quantities=quantities,
                                        scale=scale, tight=tight, show=show)
    # make the full movie
    plot.make_movie(storage, filename, progress=progress)    
   
    
    
@fill_in_docstring
def movie_multiple(storage: StorageBase,
                   filename: str,
                   quantities=None,
                   scale: ScaleData = 'automatic',
                   progress: bool = True) -> None:
    """ produce a movie for a simulation with n components

    Args:
        storage (:class:`~pse.storage.base.StorageBase`):
            The storage instance that contains all the data for the movie
        filename (str):
            The filename to which the movie is written. The extension determines
            the format used.
        quantities:
            {ARG_PLOT_QUANTITIES}
        scale (str, float, tuple of float):
            {ARG_PLOT_SCALE}
        progress (bool):
            Flag determining whether the progress of making the movie is shown.
    """
    plot = ScalarFieldPlot.from_storage(storage, quantities=quantities,
                                        scale=scale)
    plot.make_movie(storage, filename, progress=progress)




"""
The Swift-Hohenberg equation

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from typing import Callable

import numpy as np


from .base import PDEBase
from ..fields import ScalarField
from ..grids.boundaries.axes import BoundariesData
from ..tools.numba import nb, jit
from ..tools.docstrings import fill_in_docstring


        
class SwiftHohenbergPDE(PDEBase):
    r""" The Swift-Hohenberg equation 
    
    The mathematical definition is
    
    .. math::
        \partial_t c = 
            \left[\epsilon - \left(k_c^2 + \nabla^2\right)^2\right] c
            + \delta \, c^2 - c^3
        
    where :math:`c` is a scalar field and :math:`\epsilon`, :math:`k_c^2`, and
    :math:`\delta` are parameters of the equation.
    """

    explicit_time_dependence = False


    @fill_in_docstring
    def __init__(self,
                 rate: float = .1,
                 kc2: float = 1.,
                 delta: float = 1.,
                 bc: BoundariesData = 'natural', 
                 bc_lap: BoundariesData = None):
        r""" 
        Args:
            rate (float):
                The bifurcation parameter :math:`\epsilon`
            kc2 (float):
                Squared wave vector :math:`k_c^2` of the linear instability
            delta (float):
                Parameter :math:`\delta` of the non-linearity
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            bc_lap:
                The boundary conditions applied to the second derivative of the
                scalar field :math:`c`. If `None`, the same boundary condition
                as `bc` is chosen. Otherwise, this supports the same options as
                `bc`.
        """
        super().__init__()
        
        self.rate = rate
        self.kc2 = kc2
        self.delta = delta
        self.bc = bc
        self.bc_lap = bc if bc_lap is None else bc_lap
            
            
    def evolution_rate(self, state: ScalarField,  # type: ignore
                       t: float = 0) -> ScalarField:
        """ evaluate the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.ScalarField`):
                The scalar field describing the concentration distribution
            t (float): The current time point
            
        Returns:
            :class:`~pde.fields.ScalarField`:
            Scalar field describing the evolution rate of the PDE 
        """
        assert isinstance(state, ScalarField)
        state_laplace = state.laplace(bc=self.bc)
        state_laplace2 = state_laplace.laplace(bc=self.bc_lap)
        
        result = ((self.rate - self.kc2**2) * state
                  - 2 * self.kc2 * state_laplace
                  - state_laplace2
                  + self.delta * state**2 - state**3)
        result.label = 'evolution rate'
        return result  # type: ignore
     
     
    def _make_pde_rhs_numba(self, state: ScalarField  # type: ignore
                            ) -> Callable:
        """ create a compiled function evaluating the right hand side of the PDE
          
        Args:
            state (:class:`~pde.fields.ScalarField`):
                An example for the state defining the grid and data types
                  
        Returns:
            A function with signature `(state_data, t)`, which can be called
            with an instance of :class:`numpy.ndarray` of the state data and
            the time to obtained an instance of :class:`numpy.ndarray` giving
            the evolution rate.  
        """
        shape = state.grid.shape
        arr_type = nb.typeof(np.empty(shape, dtype=np.double))
        signature = arr_type(arr_type, nb.double)
          
        rate = self.rate
        kc2 = self.kc2
        delta = self.delta
        
        laplace = state.grid.get_operator('laplace', bc=self.bc)
        laplace2 = state.grid.get_operator('laplace', bc=self.bc_lap)
  
        @jit(signature)
        def pde_rhs(state_data: np.ndarray, t: float):
            """ compiled helper function evaluating right hand side """ 
            state_laplace = laplace(state_data)
            state_laplace2 = laplace2(state_laplace)
              
            return ((rate - kc2**2) * state_data
                    - 2 * kc2 * state_laplace
                    - state_laplace2
                    + delta * state_data**2 - state_data**3)
              
        return pde_rhs  # type: ignore
      



'''
Grids define the domains on which the pde will be solved. In particular, 
symmetries, periodicities, and the discretizations are also defined here.

We only consider regular, orthogonal grids, which are constructed from
orthogonal coordinate systems with equidistant discretizations along each axis.
The dimension of the space that the grid describes is given by the attribute
:attr:`dim`. Points given in these coordinates can be mapped to coordinates in
Cartesian space using the methods :meth:`point_to_cartesian` and its inverse.
Moreover, points can be mapped to cell indices using the methods
:meth:`point_to_cell`.

.. autosummary::
   :nosignatures:

   ~cartesian.UnitGrid
   ~cartesian.CartesianGrid
   ~spherical.PolarGrid
   ~spherical.SphericalGrid
   ~cylindrical.CylindricalGrid


Inheritance structure of the classes:

.. inheritance-diagram:: pde.grids.base pde.grids.cartesian pde.grids.spherical
        pde.grids.cylindrical
   :parts: 1

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
 
'''

from .base import DomainError, PeriodicityError
from .cartesian import UnitGrid, CartesianGrid
from .spherical import PolarGrid, SphericalGrid
from .cylindrical import CylindricalGrid

# import all operator modules to register the operators
from . import operators
del operators



"""
Defines a PDE class whose right hand side is given as a string

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

from collections import OrderedDict, defaultdict
from typing import Callable, Dict, Any  # @UnusedImport

import numpy as np


from pde.pdes.base import PDEBase
from pde.fields import FieldCollection
from pde.fields.base import FieldBase, DataFieldBase, OptionalArrayLike
from pde.grids.boundaries.axes import BoundariesData
from pde.tools.numba import nb, jit
from pde.tools.docstrings import fill_in_docstring


     
class PDE(PDEBase):
    """ PDE defined by a mathematical expression
    
    Attributes:
        variables (tuple):
            The name of the variables (i.e., fields) in the order they are
            expected to appear in the `state`.
        operators (tuple):
            The name of the operators that need to be defined for the expression
            to be useful. These should only contain operators defined in the
            :mod:`pde` package.
        diagnostics (dict): 
            Additional diagnostic information that might help with analyzing
            problems, e.g., when :mod:`sympy` cannot parse or :mod`numba` cannot
            compile a function.
    """
    
    @fill_in_docstring
    def __init__(self,
                 rhs: "OrderedDict[str, str]",
                 noise: OptionalArrayLike = 0,
                 bc: BoundariesData = 'natural',
                 bc_ops: Dict[str, BoundariesData] = None):
        """
        Warning:
            {WARNING_EXEC}
        
        Args:
            rhs (OrderedDict):
                The expressions defining the evolution rate. The dictionary keys
                define the name of the fields whose evolution is considered,
                while the values specify their evolution rate as a string that
                can be parsed by :mod:`sympy`. These expression may contain the
                fields themselves, standard local mathematical operators defined 
                by sympy, and the operators defined in the :mod:`pde` package.
                Note that operators need to be specified with their full name,
                i.e., `laplace` for a scalar Laplacian and `vector_laplace` for
                a Laplacian operating on a vector field.
            noise (float or :class:`numpy.ndarray`):
                Magnitude of additive Gaussian white noise. The default value of
                zero implies deterministic partial differential equations will
                be solved. Different noise magnitudes can be supplied for each
                field in coupled PDEs.
            bc:
                Boundary conditions for the operators used in the expression.
                The conditions here are applied to all operators that do not
                have a specialized condition given in `bc_ops`
                {ARG_BOUNDARIES} 
            bc_ops (dict):
                Special boundary conditions for some operators. The keys in this
                dictionary specify the operators appearing in `rhs` and the
                associated values support the same boundary conditions specified
                in `bc`. Operators that are not specified in `bc_ops` will use
                the conditions specified by `bc`
                
        Note:
            The order in which the fields are given in `rhs` defines the order
            in which they need to appear in the `state` variable when the
            evolution rate is calculated. Note that the insertion order of
            `dict` is guaranteed since Python version 3.7, so a normal
            dictionary can be used to define the equations.
        """
        from ..tools.expressions import ScalarExpression
        from sympy.core.function import AppliedUndef
        
        super().__init__(noise=noise)
        
        # validate input
        if not isinstance(rhs, OrderedDict):
            rhs = OrderedDict(rhs)
        if 't' in rhs:
            raise RuntimeError('`t` is not allowed as a variable since it '
                               'denotes time')
        
        # turn the expression strings into sympy expressions
        signature = tuple(rhs.keys()) + ('t',)
        self._rhs_expr, operators = {}, set()
        explicit_time_dependence = False
        for var, rhs_str in rhs.items():
            rhs_expr = ScalarExpression(rhs_str, signature=signature)
            self._rhs_expr[var] = rhs_expr
            
            if rhs_expr.depends_on('t'):
                explicit_time_dependence = True
            
            for func in rhs_expr._sympy_expr.atoms(AppliedUndef):
                operators.add(func.__class__.__name__)
                
        
        # set public instance attributes
        self.rhs = rhs
        self.variables = tuple(rhs.keys())
        self.operators = tuple(sorted(operators))
        self.explicit_time_dependence = explicit_time_dependence
        self.bcs: Dict[str, BoundariesData] = defaultdict(lambda: bc)
        if bc_ops is not None:
            self.bcs.update(bc_ops)
        self.diagnostics = {
                        'variables': self.variables,
                        'function_signature': signature,
                        'explicit_time_dependence': explicit_time_dependence,
                        'operators_detected': self.operators}
        self._cache: Dict[str, Any] = {}
            
        
    @property
    def expressions(self) -> Dict[str, str]:
        """ show the expressions of the PDE """
        return {k: v.expression for k, v in self._rhs_expr.items()}
            
           
    def _prepare(self, state: FieldBase) -> None:
        """ prepare the expression by setting internal variables in the cache

        Note that the expensive calculations in this method are only carried
        out if the state attributes change.

        Args:
            state (:class:`~pde.fields.FieldBase`):
                The field describing the state of the PDE
        """
        # check whether this function actually needs to be called
        if ('state_attributes' in self._cache and
                state.attributes == self._cache['state_attributes']):
            return  # prepare was already called
        self._cache = {}  # clear cache, if there was something

        # check whether the state is compatible with the PDE
        num_fields = len(self.variables)
        self.diagnostics['num_fields'] = num_fields
        if isinstance(state, FieldCollection):
            if num_fields != len(state):
                raise ValueError(f'Expected {num_fields} fields in state, but '
                                 f'got {len(state)} ones')
        elif isinstance(state, DataFieldBase):
            if num_fields != 1:
                raise ValueError(f'Expected {num_fields} fields in state, but '
                                 'got only one')
        else:
            raise ValueError(f'Unknown state class {state.__class__.__name__}')
        
        # obtain the (differential) operators
        user_funcs = {}
        for func in self.operators:
            user_funcs[func] = state.grid.get_operator(func, bc=self.bcs[func])
            
        # obtain the python functions for the rhs
        self._cache['rhs_funcs'] = [self._rhs_expr[var].
                                        _get_function(user_funcs=user_funcs)
                                    for var in self.variables]

        # add extra information for field collection
        if isinstance(state, FieldCollection):
            # isscalar be False even if start == stop (e.g. vector fields)
            isscalar = tuple(field.rank == 0 for field in state)  
            starts = tuple(slc.start for slc in state._slices)
            stops = tuple(slc.stop for slc in state._slices)

            def get_data_tuple(state_data):
                """ helper for turning state_data into a tuple of field data """
                return tuple(state_data[starts[i]]
                                 if isscalar[i]
                                 else state_data[starts[i]: stops[i]]
                             for i in range(num_fields))
            self._cache['get_data_tuple'] = get_data_tuple

        # store the attributes in the cache, which allows to later circumvent
        # calculating the quantities above again. Note that this has to be the
        # last expression of the method, so the cache is only valid when the
        # prepare function worked successfully            
        self._cache['state_attributes'] = state.attributes
        
        
    def evolution_rate(self, state: FieldBase, t: float = 0) -> FieldBase:
        """ evaluate the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                The field describing the state of the PDE
            t (float):
                The current time point
            
        Returns:
            :class:`~pde.fields.FieldBase`:
            Field describing the evolution rate of the PDE 
        """
        self._prepare(state)
            
        if isinstance(state, DataFieldBase):
            rhs = self._cache['rhs_funcs'][0]
            return state.copy(data=rhs(state.data, t))
        
        elif isinstance(state, FieldCollection):
            result = state.copy()
            for i in range(len(state)):
                data_tpl = self._cache['get_data_tuple'](state.data)
                result[i].data = self._cache['rhs_funcs'][i](*data_tpl, t)
            return result
        
        else:
            raise TypeError(f'Unsupported field {state.__class__.__name__}')
    
    
    def _make_pde_rhs_numba_coll(self, state: FieldCollection) -> Callable:
        """ create the compiled rhs if `state` is a field collection
        
        Args:
            state (:class:`~pde.fields.FieldCollection`):
                An example for the state defining the grid and data types
                
        Returns:
            A function with signature `(state_data, t)`, which can be called
            with an instance of :class:`numpy.ndarray` of the state data and
            the time to obtained an instance of :class:`numpy.ndarray` giving
            the evolution rate.
        """
        num_fields = len(state)
        data_shape = state.data.shape
        rhs_list = tuple(jit(self._cache['rhs_funcs'][i])
                         for i in range(num_fields))

        starts = tuple(slc.start for slc in state._slices)
        stops = tuple(slc.stop for slc in state._slices)

        # In the future, the following should be possible:
#         @jit
#         def evolution_rate(state_data, t, out):
#             """ evolve all agents explicitly """
#             for i in nb.literal_unroll(range(num_fields)):
#                 out[i] = rhs_list[i](*state_data, t)
#         return evolver
    
        get_data_tuple = self._cache['get_data_tuple']
            
        def chain(i=0, inner=None):
            """ recursive helper function for applying all rhs """
            # run through all evolvers
            rhs = rhs_list[i]

            if inner is None:
                # the innermost function does not need to call a child
                @jit
                def wrap(data_tpl, t, out):
                    out[starts[i]: stops[i]] = rhs(*data_tpl, t)

            else:
                # all other functions need to call one deeper in the chain
                @jit
                def wrap(data_tpl, t, out):
                    inner(data_tpl, t, out)
                    out[starts[i]: stops[i]] = rhs(*data_tpl, t)
            
            if i < num_fields - 1:
                # there are more items in the chain
                return chain(i + 1, inner=wrap)
            else:
                # this is the outermost function
                @jit
                def evolution_rate(state_data: np.ndarray, t: float = 0):
                    out = np.empty(data_shape)
                    with nb.objmode():
                        data_tpl = get_data_tuple(state_data)
                        wrap(data_tpl, t, out)
                    return out
                return evolution_rate
        
        # compile the recursive chain
        return chain()  # type: ignore            
    
    
    def _make_pde_rhs_numba(self, state: FieldBase) -> Callable:
        """ create a compiled function evaluating the right hand side of the PDE
        
        Args:
            state (:class:`~pde.fields.FieldBase`):
                An example for the state defining the grid and data types
                
        Returns:
            A function with signature `(state_data, t)`, which can be called
            with an instance of :class:`numpy.ndarray` of the state data and
            the time to obtained an instance of :class:`numpy.ndarray` giving
            the evolution rate.
        """
        self._prepare(state)
        
        if isinstance(state, DataFieldBase):
            return jit(self._cache['rhs_funcs'][0])  # type: ignore
        elif isinstance(state, FieldCollection):
            return self._make_pde_rhs_numba_coll(state)
        else:
            raise TypeError(f'Unsupported field {state.__class__.__name__}')
        


'''
Tools for plotting and controlling plot output using context managers

.. autosummary::
   :nosignatures:

   finalize_plot
   disable_interactive
   plot_on_axes
   plot_on_figure
   PlotReference
   BasicPlottingContext
   JupyterPlottingContext
   get_plotting_context
   napari_viewer

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import re
import contextlib
import logging
import warnings
from typing import Type, Dict, Tuple, Any, TYPE_CHECKING  # @UnusedImport

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.axes as mpl_axes

if TYPE_CHECKING:
    from ..grids.base import GridBase  # @UnusedImport

 
 
_IS_PLOTTING = False  # global flag to detect nested plotting calls



def finalize_plot(fig_or_ax=None,
                  title: str = None,
                  filename: str = None,
                  show: bool = False,
                  close_figure: bool = False) -> Tuple[Any, Any]:
    r""" finalizes a figure by adjusting relevant parameters
    
    Args:
        fig_or_ax:
            The figure or axes that are affected. If `None`, the current figure
            is used.
        title (str):
            Determines the title of the figure
        filename (str):
            If given, the resulting image is written to this file.
        show (bool):
            Flag determining whether :func:`matplotlib.pyplot.show` is called
        close_figure (bool):
            Whether the figure should be closed in the end
            
    Returns:
        tuple: The figure and the axes that were used to finalize the plot
    """
    # determine which figure to modify    
    if fig_or_ax is None:
        fig = plt.gcf()  # current figure
        ax = fig.gca()
    elif hasattr(fig_or_ax, 'savefig'):
        fig = fig_or_ax  # figure is given
        ax = fig.gca()
    else:
        ax = fig_or_ax  # assume that axes are given
        fig = ax.get_figure()
    
    if title is not None:
        ax.set_title(title)
    
    if filename:
        fig.savefig(filename)
    if show:
        plt.show()
    if close_figure:
        plt.close(fig)
        
    return fig, ax
    
    

@contextlib.contextmanager
def disable_interactive():
    """ context manager disabling the interactive mode of matplotlib
    
    This context manager restores the previous state after it is done. Details
    of the interactive mode are described in :func:`matplotlib.interactive`.
    """
    if plt.isinteractive():
        # interactive mode is enabled => disable it temporarily
        plt.interactive(False) 
        yield
        plt.interactive(True)
        
    else:
        # interactive mode is already disabled => do nothing
        yield



class plot_on_axes:
    """ wrapper for a plot method that fills an axes
    
    Example:
        The following example illustrates how the classes in this module can be
        used to implement plotting on a class. In particular, supplying the
        `update_method` will allow efficient dynamical plotting::
    
            class State:
                def __init__(self):
                    self.data = np.arange(8)
    
                def _update_plot(self, reference):
                    reference.element.set_ydata(self.data)
    
                @plot_on_axes(update_method='_update_plot')
                def plot(self, ax):
                    line, = ax.plot(np.arange(8), self.data)
                    return PlotReference(ax, line)
    """
    
    
    def __init__(self, update_method=None):
        """
        Args:
            update_method (callable):
                Method to call to update the plot. The argument of the new
                method will be the result of the initial call of the wrapped
                method.
        """
        # initialize the decorator
        if callable(update_method):
            raise RuntimeError('Wrapper needs to be called with brackets')
        self.update_method = update_method
        
        
    def __call__(self, method):
        """ apply the actual decorator """
        
        def wrapper(obj, *,
                    title: str = None,
                    filename: str = None,
                    show: bool = None,
                    close_figure: bool = False,
                    ax=None,
                    **kwargs):
            """
            Args:
                title (str):
                    Title of the plot. If omitted, the title might be chosen
                    automatically.
                filename (str, optional):
                    If given, the plot is written to the specified file.
                show (bool):
                    Flag setting whether :func:`matplotlib.pyplot.show` is
                    called. The value `None` sets show to `True` by default, but
                    disables it for nested calls.                    
                close_figure (bool):
                    Flag setting whether the figure is closed (after it was
                    shown)
            """
            # some logic to check for nested plotting calls:
            global _IS_PLOTTING
            root_plotting_call = not _IS_PLOTTING
            if show is None:
                show = root_plotting_call  # only call show outer routine
            _IS_PLOTTING = True
            
            # disable interactive plotting temporarily
            with disable_interactive():
                
                if ax is None:
                    # create new figure
                    backend = mpl.get_backend()
                    if 'backend_inline' in backend or 'nbAgg' == backend:
                        plt.close('all')  # close figures that were left over
                    fig, ax = plt.subplots()
                else:
                    fig = ax.get_figure()
            
                # call the actual plotting function
                reference = method(obj, ax=ax, **kwargs)
                
                # finishing touches...            
                if title is not None:
                    reference.ax.set_title(title)
                if filename:
                    fig.savefig(filename)
                    
            # decide what to do with the final plot
            if show:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    plt.show()
            if close_figure:
                plt.close(fig)
                
            if root_plotting_call:
                _IS_PLOTTING = False
                
            return reference
            
        wrapper.__name__ = method.__name__        
        wrapper.__module__ = method.__module__
        
        doc_fragment = wrapper.__doc__.splitlines()[:-1]
        doc_fragment = '\n'.join(line[4:] for line in doc_fragment)
        wrapper.__doc__ = re.sub(r'Args:\s*$', doc_fragment, method.__doc__,
                                 flags=re.MULTILINE)
        wrapper.__dict__.update(method.__dict__)

        wrapper.mpl_class = 'axes'
        wrapper.update_method = self.update_method
            
        return wrapper
    


class plot_on_figure:
    """ wrapper for a plot method that fills an entire figure """
    
    
    def __init__(self, update_method=None):
        """
        Args:
            update_method (callable):
                Method to call to update the plot. The argument of the new
                method will be the result of the initial call of the wrapped
                method.
        """
        # initialize the decorator
        if callable(update_method):
            raise RuntimeError('Wrapper needs to be called with brackets')
        self.update_method = update_method
        
        
    def __call__(self, method):
        """ apply the actual decorator """
        
        def wrapper(obj, *,
                    title: str = None,
                    constrained_layout: bool = True,
                    filename: str = None,
                    show: bool = None,
                    close_figure: bool = False,
                    fig=None,
                    **kwargs):
            """
            Args:
                title (str):
                    Title of the plot. If omitted, the title might be chosen
                    automatically. This is shown above all panels.
                constrained_layout (bool):
                    Whether to use `constrained_layout` in 
                    :func:`matplotlib.pyplot.figure` call to create a figure. 
                    This affects the layout of all plot elements.
                filename (str, optional):
                    If given, the figure is written to the specified file.
                show (bool):
                    Flag setting whether :func:`matplotlib.pyplot.show` is
                    called. The value `None` sets `show` to `True` by default,
                    but disables it for nested calls.                  
                close_figure (bool):
                    Flag setting whether the figure is closed (after it was
                    shown).
            """
            # some logic to check for nested plotting calls:
            global _IS_PLOTTING
            root_plotting_call = not _IS_PLOTTING
            if show is None:
                show = root_plotting_call  # only call show outer routine
            _IS_PLOTTING = True

            # disable interactive plotting temporarily
            with disable_interactive():
                
                if fig is None:
                    # create new figure
                    backend = mpl.get_backend()
                    if 'backend_inline' in backend or 'nbAgg' == backend:
                        plt.close('all')  # close figures that were left over
                    fig = plt.figure(constrained_layout=constrained_layout)
            
                # call the actual plotting function
                reference = method(obj, fig=fig, **kwargs)
                
                # finishing touches...            
                if title is not None:
                    fig.suptitle(title)
                if filename:
                    fig.savefig(filename)
                    
            # decide what to do with the final plot
            if show:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    plt.show()
                    
            if close_figure:
                plt.close(fig)    
                
            if root_plotting_call:
                _IS_PLOTTING = False
                
            return reference    
            
        wrapper.__name__ = method.__name__        
        wrapper.__module__ = method.__module__
        
        doc_fragment = wrapper.__doc__.splitlines()[:-1]
        doc_fragment = '\n'.join(line[4:] for line in doc_fragment)
        wrapper.__doc__ = re.sub(r'Args:\s*$', doc_fragment, method.__doc__,
                                 flags=re.MULTILINE)
        wrapper.__dict__.update(method.__dict__)

        wrapper.mpl_class = 'figure'
        wrapper.update_method = self.update_method
        
        return wrapper



class PlotReference():
    """ contains all information to update a plot element """
    
    __slots__ = ['ax', 'element', 'parameters']
    
    def __init__(self, ax, element: Any, parameters: Dict[str, Any] = None):
        """
        Args:
            ax (:class:`matplotlib.axes.Axes`): The axes of the element
            element (:class:`matplotlib.artist.Artist`): The actual element 
            parameters (dict): Parameters to recreate the plot element
        """
        self.ax = ax
        self.element = element
        self.parameters = {} if parameters is None else parameters



class PlottingContextBase(object):
    """ base class of the plotting contexts
    
    Example:
        The context wraps calls to the :mod:`matplotlib.pyplot` interface::
    
            context = PlottingContext()
            with context:
                plt.plot(...)
                plt.xlabel(...)
    """
    
    supports_update: bool = True
    """ flag indicating whether the context supports that plots can be updated
    with out redrawing the entire plot """
    
    
    def __init__(self,
                 title: str = None,
                 show: bool = True):
        """ 
        Args:
            title (str): The shown in the plot
            show (bool): Flag determining whether plots are actually shown
        """
        self.title = title
        self.show = show
        
        self.initial_plot = True
        self.fig = None
        self._logger = logging.getLogger(__name__)
        self._logger.info(f'Initialize {self.__class__.__name__}')
    

    def __enter__(self):
        # start the plotting process
        if self.fig is not None:
            plt.figure(self.fig.number)
    

    def __exit__(self, *exc): 
        if self.initial_plot or not self.supports_update:
            # recreate the entire figure
            self.fig = plt.gcf()
            if len(self.fig.axes) == 0:
                # The figure seems to be empty, which must be a mistake
                raise RuntimeError('Plot figure does not contain axes')
            
            elif len(self.fig.axes) == 1:
                # The figure contains only a single axis, indicating that it is
                # composed of a single panel
                self._title = plt.title(self.title)
                
            else:
                # The figure contains multiple axes. This is an indication that
                # the figure consists of multiple panels, although insets and
                # colorbars also count as additional axes
                self._title = plt.suptitle(self.title)
                
            self.initial_plot = False
            
        else:
            # update the old figure
            self._title.set_text(self.title)


    def close(self):
        """ close the plot """
        # close matplotlib figure
        if self.fig is not None:
            plt.close(self.fig)



class BasicPlottingContext(PlottingContextBase):
    """ basic plotting using just matplotlib """
    
    def __init__(self,
                 fig_or_ax=None,
                 title: str = None,
                 show: bool = True):
        """
        Args:
            fig_or_ax:
                If axes are given, they are used. If a figure is given, it is 
                set as active.
            title (str):
                The shown in the plot
            show (bool):
                Flag determining whether plots are actually shown
        """
        super().__init__(title=title, show=show)
        
        # determine which figure to modify
        if isinstance(fig_or_ax, mpl_axes.Axes):
            self.fig = fig_or_ax.get_figure()  # assume that axes are given
        elif isinstance(fig_or_ax, mpl.figure.Figure):
            self.fig = fig_or_ax
    

    def __exit__(self, *exc): 
        super().__exit__(*exc)
        if self.show:
            self.fig.canvas.draw()  # required for display in nbagg backend
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # add a small pause to allow the GUI to run it's event loop
                plt.pause(1e-3)



class JupyterPlottingContext(PlottingContextBase):
    """ plotting in a jupyter widget using the `inline` backend """
    
    supports_update = False
    """ flag indicating whether the context supports that plots can be updated
    with out redrawing the entire plot. The jupyter backend (`inline`) requires
    replotting of the entire figure, so an update is not supported."""
    
    def __enter__(self):
        from IPython.display import display
        from ipywidgets import Output
        
        if self.initial_plot:
            # close all previous plots
            plt.close('all')
            
            # create output widget for capturing all plotting
            self._ipython_out = Output()
            
            if self.show:
                # only show the widget if necessary
                display(self._ipython_out)
            
        # capture plots in the output widget
        self._ipython_out.__enter__()
    
    
    def __exit__(self, *exc):
        # finalize plot
        super().__exit__(*exc)
        
        if self.show:
            # show the plot, but ...
            plt.show()  # show the figure to make sure it can be captured
        # ... also clear it the next time something is done        
        self._ipython_out.clear_output(wait=True)
            
        # stop capturing plots in the output widget
        self._ipython_out.__exit__(*exc)

        # close the figure, so figure windows do not accumulate
        plt.close(self.fig)
 

    def close(self):
        """ close the plot """
        super().close()
        # close ipython output
        try:
            self._ipython_out.close()
        except Exception:
            pass


        
def get_plotting_context(context=None,
                         title: str = None,
                         show: bool = True) -> PlottingContextBase:
    """ returns a suitable plotting context
    
    Args:
        context:
            An instance of :class:`PlottingContextBase` or an instance of
            :class:`matplotlib.axes.Axes` or :class:`matplotlib.figure.Figure`
            to determine where the plotting will happen. If omitted, the context
            is determined automatically.
        title (str):
            The title shown in the plot
        show (bool):
            Determines whether the plot is shown while the simulation is
            running. If `False`, the files are created in the background.
            
    Returns:
        :class:`PlottingContextBase`: The plotting context
    """
    if context is None:
        # figure out whether plots are shown in jupyter notebook
        
        if 'backend_inline' in mpl.get_backend():
            # special context to support the `inline` backend
            try:
                from IPython.display import display  # @UnusedImport
                from ipywidgets import Output  # @UnusedImport
            except ImportError:
                context_class: Type[PlottingContextBase] = BasicPlottingContext
            else:
                context_class = JupyterPlottingContext
                
        else:
            # standard context for all other backends
            context_class = BasicPlottingContext
        
        return context_class(title=title, show=show)    
    
    elif isinstance(context, PlottingContextBase):
        # re-use an existing context
        context.title = title
        context.show = show
        return context
    
    elif isinstance(context, (mpl_axes.Axes, mpl.figure.Figure)):
        # create a basic context based on the given axes or figure
        return BasicPlottingContext(fig_or_ax=context, title=title, show=show)
    
    else:
        raise RuntimeError(f'Unknown plotting context `{context}`')
        
        

@contextlib.contextmanager
def napari_viewer(grid: 'GridBase', **kwargs):
    """ creates an napari viewer for interactive plotting
    
    Args:
        grid (:class:`pde.grids.base.GridBase`): The grid defining the space
        **kwargs: Extra arguments are passed to :class:`napari.Viewer`
    """
    import napari
    
    if grid.num_axes == 1:
        raise RuntimeError('Interactive plotting only works for data with '
                           'at least 2 dimensions')
    
    viewer_args = kwargs
    viewer_args.setdefault('axis_labels', grid.axes)
    viewer_args.setdefault('ndisplay', 3 if grid.num_axes >= 3 else 2)
    
    with napari.gui_qt():  # create Qt GUI context
        yield napari.Viewer(**viewer_args)        
        


'''
Defines a scalar field over a grid

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

from typing import Union, Sequence, Dict, Optional, TYPE_CHECKING
from pathlib import Path

import numpy as np

from .base import DataFieldBase
from ..grids import UnitGrid, CartesianGrid
from ..grids.base import GridBase, DomainError
from ..tools.docstrings import fill_in_docstring


if TYPE_CHECKING:
    from ..grids.boundaries.axes import BoundariesData  # @UnusedImport
    from .vectorial import VectorField  # @UnusedImport



class ScalarField(DataFieldBase):
    """ Single scalar field on a grid

    Attributes:
        grid (:class:`~pde.grids.GridBase`):
            The underlying grid defining the discretization
        data (:class:`np.ndarray`):
            Scalar values at the support points of the grid
        label (str):
            Name of the field
    """

    rank = 0
        
        
    @classmethod
    @fill_in_docstring
    def from_expression(cls, grid: GridBase, expression: str,
                        label: str = None) -> "ScalarField":
        """ create a scalar field on a grid from a given expression
        
        Warning:
            {WARNING_EXEC}
        
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            expression (str):
                Mathematical expression for the scalar value as a function of
                the position on the grid. The expression may contain standard
                mathematical functions and it may depend on the axes labels of
                the grid.
            label (str, optional):
                Name of the field
        """
        from ..tools.expressions import ScalarExpression
        expr = ScalarExpression(expression=expression, signature=grid.axes)
        points = {name: grid.cell_coords[..., i]
                  for i, name in enumerate(grid.axes)}
        return cls(grid=grid,  # lgtm [py/call-to-non-callable]
                   data=expr(**points),
                   label=label)
    
    
    @classmethod
    def from_image(cls, path: Union[Path, str], bounds=None, periodic=False,
                   label: str = None) -> "ScalarField":
        """ create a scalar field from an image
    
        Args:
            path (:class:`Path` or str): The path to the image
            bounds (tuple, optional): Gives the coordinate range for each axis.
                This should be two tuples of two numbers each, which mark the
                lower and upper bound for each axis. 
            periodic (bool or list): Specifies which axes possess periodic
                boundary conditions. This is either a list of booleans defining
                periodicity for each individual axis or a single boolean value
                specifying the same periodicity for all axes.
            label (str, optional):
                Name of the field
        """
        from matplotlib.pyplot import imread
        # read image and convert to grayscale
        data = imread(path)
        if data.ndim == 2:
            pass  # is already gray scale
        elif data.ndim == 3:
            # convert to gray scale using ITU-R 601-2 luma transform:
            weights = np.array([0.299, 0.587, 0.114])
            data = data[..., :3] @ weights
        else:
            raise RuntimeError(f'Image data has wrong shape: {data.shape}')
        
        # transpose data to use mathematical conventions for axes
        data = data.T[:, ::-1]
        
        # determine the associated grid
        if bounds is None:
            grid: GridBase = UnitGrid(data.shape, periodic=periodic)
        else:
            grid = CartesianGrid(bounds, data.shape, periodic=periodic)
        
        return cls(grid, data, label=label)
        
     
    @DataFieldBase._data_flat.setter  # type: ignore
    def _data_flat(self, value):
        """ set the data from a value from a collection """
        self._data = value[0]

        
    @fill_in_docstring
    def laplace(self, bc: "BoundariesData",
                out: Optional['ScalarField'] = None,
                label: str = 'laplace') -> 'ScalarField':
        """ apply Laplace operator and return result as a field 
        
        Args:
            bc:
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            out (ScalarField, optional):
                Optional scalar field to which the  result is written.
            label (str, optional):
                Name of the returned field
            
        Returns:
            ScalarField: the result of applying the operator 
        """
        if out is not None:
            assert isinstance(out, ScalarField)
        laplace = self.grid.get_operator('laplace', bc=bc)
        return self.apply(laplace, out=out, label=label)

        
    @fill_in_docstring
    def gradient(self, bc: "BoundariesData",
                 out: Optional['VectorField'] = None,
                 label: str = 'gradient') -> 'VectorField':
        """ apply gradient operator and return result as a field 
        
        Args:
            bc: 
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            out (VectorField, optional):
                Optional vector field to which the result is written.
            label (str, optional):
                Name of the returned field
            
        Returns:
            VectorField: the result of applying the operator 
        """
        from .vectorial import VectorField  # @Reimport
        
        gradient = self.grid.get_operator('gradient', bc=bc)
        if out is None:
            out = VectorField(self.grid, gradient(self.data), label=label)
        else:
            assert isinstance(out, VectorField)
            gradient(self.data, out=out.data)
        return out
    
    
    @fill_in_docstring
    def solve_poisson(self, bc: "BoundariesData",
                      out: Optional['ScalarField'] = None,
                      label: str = "Solution to Poisson's equation"):
        r""" solve Poisson's equation with the current field as inhomogeneity.
         
        Denoting the current field by :math:`x`, we thus solve for :math:`y`,
        defined by the equation 
 
        .. math::
            \nabla^2 y(\boldsymbol r) = -x(\boldsymbol r)
            
        with boundary conditions specified by `bc`.
            
        Note:
            In case of periodic or Neumann boundary conditions, the right hand
            side :math:`x(\boldsymbol r)` needs to satisfy the following
            condition for consistency:
            
            .. math::
                \int x \, \mathrm{d}V = \oint g \, \mathrm{d}S
                
            where :math:`g` denotes the function specifying the outwards
            derivative for Neumann conditions. In particular, the integral over
            :math:`x` must vanish for neutral Neumann or periodic conditions.
             
        Args:
            bc: 
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            out (ScalarField, optional):
                Optional scalar field to which the  result is written.
            label (str, optional):
                Name of the returned field
             
        Returns:
            ScalarField: the result of applying the operator 
        """
        # Deprecated this method on 2020-04-15
        import warnings
        warnings.warn("solve_poisson() method is deprecated. Use the function "
                      "pde.pdes.solve_poisson_equation or pde.pdes.solve_"
                      "laplace_equation instead.",
                      DeprecationWarning)
        
        # solve the poisson problem
        solve_poisson = self.grid.get_operator('poisson_solver', bc=bc)
        try:
            result = solve_poisson(self.data)
        except RuntimeError:
            average = self.average
            if abs(average) > 1e-10:
                raise RuntimeError('Could not solve the Poisson problem. One '
                                   'possible reason for this is that only '
                                   'periodic or Neumann conditions are '
                                   'applied although the average of the field '
                                   f'is {average} and thus non-zero.')
            else:
                raise  # another error occured
         
        if out is None:
            return ScalarField(self.grid, result, label=label)
        else:
            out.data = result
            if label:
                out.label = label
            return out
    
        
    @property
    def integral(self) -> float:
        """ float: integral of the scalar field over space """
        return float(self.grid.integrate(self.data))

               
    def project(self, axes: Union[str, Sequence[str]],
                method: str = 'integral',
                label: str = None) -> "ScalarField":
        """ project scalar field along given axes
        
        Args:
            axes (list of str):
                The names of the axes that are removed by the projection
                operation. The valid names for a given grid are the ones in
                the :attr:`GridBase.axes` attribute.
            method (str):
                The projection method. This can be either 'integral' to
                integrate over the removed axes or 'average' to perform an
                average instead.
            label (str, optional):
                The label of the returned field
                
        Returns:
            ScalarField: The projected data in a scalar field with a subgrid of
            the original grid.
        """
        if any(ax not in self.grid.axes for ax in axes):
            raise ValueError(f'The axes {axes} are not all contained in '
                             f'{self.grid} with axes {self.grid.axes}')
            
        # determine the axes after projection
        ax_all = range(self.grid.num_axes)
        ax_remove = tuple(self.grid.axes.index(ax) for ax in axes)
        ax_retain = tuple(sorted(set(ax_all) - set(ax_remove)))
        
        # determine the new grid
        subgrid = self.grid.get_subgrid(ax_retain)
        
        # calculate the new data
        if method == 'integral':
            subdata = self.grid.integrate(self.data, axes=ax_remove)
        elif method == 'average' or method == 'mean':
            subdata = (self.grid.integrate(self.data, axes=ax_remove) /
                       self.grid.integrate(1, axes=ax_remove))
        else:
            raise ValueError(f'Unknown projection method `{method}`')
        
        # create the new field instance
        return self.__class__(grid=subgrid, data=subdata, label=label)
    
    
    def slice(self, position: Dict[str, float],
              method: str = 'nearest',
              label: str = None) -> "ScalarField":
        """ slice data at a given position
        
        Args:
            position (dict):
                Determines the location of the slice using a dictionary
                supplying coordinate values for a subset of axes. Axes not
                mentioned in the dictionary are retained and form the slice.
                For instance, in a 2d Cartesian grid, `position = {'x': 1}`
                slices along the y-direction at x=1. Additionally, the special
                positions 'low', 'mid', and 'high' are supported to reference
                relative positions along the axis.
            method (str):
                The method used for slicing. `nearest` takes data from cells
                defined on the grid.
            label (str, optional):
                The label of the returned field
                
        Returns:
            ScalarField: The sliced data in a scalar field with a subgrid of
            the original grid.
        """
        grid = self.grid
        
        # parse the positions and determine the axes to remove
        ax_remove, pos_values = [], np.zeros(grid.num_axes)
        for ax, pos in position.items():
            # check the axis
            try:
                i = grid.axes.index(ax)
            except ValueError:
                raise ValueError(f'The axes {ax} is not contained in '
                                 f'{self.grid} with axes {self.grid.axes}')
            ax_remove.append(i)
            
            # check the position
            if isinstance(pos, str):
                if pos in {'min', 'low', 'lower'}:
                    pos_values[i] = grid.axes_coords[i][0]
                elif pos in {'max', 'high', 'upper'}:
                    pos_values[i] = grid.axes_coords[i][-1]
                elif pos in {'mid', 'middle', 'center'}:
                    pos_values[i] = np.mean(grid.axes_bounds[i])
                else:
                    raise ValueError(f'Unknown position `{pos}`')
            else:
                pos_values[i] = float(pos)
            
        # determine the axes left after slicing and the new grid
        ax_all = range(grid.num_axes)
        ax_retain = tuple(sorted(set(ax_all) - set(ax_remove)))
        subgrid = grid.get_subgrid(ax_retain)
        
        # obtain the sliced data
        if method == 'nearest':
            idx = []
            for i in range(grid.num_axes):
                if i in ax_remove:
                    pos = pos_values[i]
                    axis_bounds = grid.axes_bounds[i]
                    if pos < axis_bounds[0] or pos > axis_bounds[1]:
                        raise DomainError(f'Position {grid.axes[i]} = {pos} is '
                                          'outside the domain')
                    # add slice that is closest to pos 
                    idx.append(np.argmin((grid.axes_coords[i] - pos)**2))
                else:
                    idx.append(slice(None))
            subdata = self.data[tuple(idx)]
            
        else:
            raise ValueError(f'Unknown slicing method `{method}`')
    
        # create the new field instance
        return self.__class__(grid=subgrid, data=subdata, label=label)
    
        
    def to_scalar(self, scalar: str = 'auto',
                  label: Optional[str] = None) -> "ScalarField":
        """ return a modified scalar field by applying `method`
        
        Args:
            scalar (str or int):
                How to obtain the scalar. For ScalarField, the default `0`
                simply returns the actual field. Setting this to 'norm' returns
                the absolute value at each point.
            label (str, optional): Name of the returned field
            
        Returns:
            :class:`pde.fields.scalar.ScalarField`: the scalar field after
            applying the operation
        """
        if scalar == 'auto':
            data = self.data
            
        elif scalar == 'abs' or scalar == 'norm':
            data = np.abs(self.data)
            
        elif scalar == 'squared_sum':
            data = self.data**2
            
        else:
            raise ValueError(f'Unknown method `{scalar}` for `to_scalar`')
        
        return ScalarField(grid=self.grid, data=data, label=label)




"""
Base classes for storing data
   
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de> 
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Optional, List, Tuple, Iterator, Union, Sequence, Any

import numpy as np

from ..grids.base import GridBase
from ..fields import ScalarField, VectorField, Tensor2Field
from ..fields.base import FieldBase
from ..trackers.base import TrackerBase, InfoDict
from ..trackers.intervals import IntervalType, IntervalData
from ..tools.docstrings import fill_in_docstring


    
class StorageBase(metaclass=ABCMeta):
    """ base class for storage of discretized fields """
    
    
    times: Sequence[float]  # :class:`numpy.ndarray`): stored time points
    data: Any  # The actual data for all the stored times
    

    def __init__(self, info: InfoDict = None,
                 write_mode: str = 'truncate_once'):
        """
        Args:
            info (dict):
                Supplies extra information that is stored in the storage
            write_mode (str):
                Determines how new data is added to already existing one.
                Possible values are: 'append' (data is always appended),
                'truncate' (data is cleared every time this storage is used
                for writing), or 'truncate_once' (data is cleared for the first
                writing, but subsequent data using the same instances are
                appended). Alternatively, specifying 'readonly' will disable
                writing completely.
        """
        self.info = {} if info is None else info
        self.write_mode = write_mode
        self._data_shape: Optional[Tuple[int, ...]] = None
        self._grid: Optional[GridBase] = None
        self._field: Optional[FieldBase] = None
        self._logger = logging.getLogger(self.__class__.__name__)
    
    
    @property
    def data_shape(self) -> Tuple[int, ...]:
        """ the current data shape.
        
        Raises:
            RuntimeError: if data_shape was not set
        """
        if self._data_shape is None:
            raise RuntimeError('data_shape was not set')
        else:  # use the else clause to help typing
            return self._data_shape


    @abstractmethod
    def append(self, data: np.ndarray, time: Optional[float] = None) -> None:
        pass
    
    
    def clear(self, clear_data_shape: bool = False) -> None:
        """ truncate the storage by removing all stored data.
        
        Args:
            clear_data_shape (bool): Flag determining whether the data shape is
                also deleted.
        """
        if clear_data_shape:
            self._data_shape = None
    

    def __len__(self):
        """ return the number of stored items, i.e., time steps """
        return len(self.times)
        

    @property
    def shape(self) -> Optional[Tuple[int, ...]]:
        """ the shape of the stored data """
        if self._data_shape:
            return (len(self),) + self._data_shape 
        else:
            return None
    
    
    @property
    def grid(self) -> Optional[GridBase]:
        """ GridBase: the grid associated with this storage
        
        This returns `None` if grid was not stored in `self.info`.
        """
        if self._grid is None:
            if 'field_attributes' in self.info:
                attrs_serialized = self.info['field_attributes']
                attrs = FieldBase.unserialize_attributes(attrs_serialized)
                self._grid = attrs['grid']
            else:
                self._logger.warning('`grid` attribute was not stored')
        return self._grid
    
    
    def get_field(self, index: int) -> FieldBase:
        """ return the field corresponding to index
        
        Load the data associated with a given index, i.e., with time 
        `self.times[index]`.
        
        Args:
            index (int):
                The index of the data to load
                
        Returns:
            :class:`~pde.fields.FieldBase`:
            The field class containing the grid and data
        """
        if self._field is None:
            # we need to determine the field type
            
            if self.grid is None:
                raise RuntimeError('Could not load grid from data. Please set '
                                   'the `_grid` attribute to the grid that has '
                                   'been used for the stored data.')
            
            if 'field_attributes' in self.info:
                # field type was stored in data
                attrs_serialized = self.info['field_attributes']
                attrs = FieldBase.unserialize_attributes(attrs_serialized)
                self._field = FieldBase.from_state(attrs)
                
            else:
                # try to determine field type automatically

                # obtain data shape by removing the first axis (associated with
                # the time series and the last axes (associated with the spatial
                # dimensions). What is left should be the (local) data stored
                # at each grid point for each time step. Note that self.data
                # might be a list of arrays
                local_shape = self.data[index].shape[:-self.grid.num_axes]
                dim = self.grid.dim
                if len(local_shape) == 0:  # rank 0
                    self._field = ScalarField(self.grid)
                elif local_shape == (dim,):  # rank 1
                    self._field = VectorField(self.grid)
                elif local_shape == (dim, dim):  # rank 2
                    self._field = Tensor2Field(self.grid)
                else:
                    raise RuntimeError('`field` attribute was not stored in '
                                       f'file and the data shape {local_shape} '
                                       'could not be interpreted automatically')
                self._logger.warning('`field` attribute was not stored. We '
                                     'guessed that the data is of type '
                                     f'{self._field.__class__.__name__}.')
                    
        # create the field with the data of the given index
        return self._field.copy(data=self.data[index])
        
        
    def __getitem__(self, key: Union[int, slice]) \
            -> Union[FieldBase, List[FieldBase]]:
        """ return field at given index or a list of fields for a slice """
        if isinstance(key, int):
            return self.get_field(key)
        elif isinstance(key, slice):
            return [self.get_field(i)
                    for i in range(*key.indices(len(self)))]
        else:
            raise TypeError('Unknown key type')
    
    
    def __iter__(self) -> Iterator[FieldBase]:
        """ iterate over all stored fields """
        for i in range(len(self)):
            yield self[i]  # type: ignore
    
    
    def items(self) -> Iterator[Tuple[float, FieldBase]]:
        """ iterate over all times and stored fields, returning pairs """
        for i in range(len(self)):
            yield self.times[i], self[i]  # type: ignore

    
    @fill_in_docstring
    def tracker(self, interval: Union[int, float, IntervalType] = 1) \
            -> "StorageTracker":
        """ create object that can be used as a tracker to fill this storage
        
        Args:
            interval:
                {ARG_TRACKER_INTERVAL}
            
        Returns:
            :class:`~pde.trackers.trackers.StorageTracker`
            The tracker that fills the current storage
        """
        return StorageTracker(storage=self, interval=interval)
    
    
    def start_writing(self, field: FieldBase, info: InfoDict = None) -> None:
        """ initialize the storage for writing data
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be written to extract the grid
                and the data_shape
            info (dict):
                Supplies extra information that is stored in the storage
        """
        if self.write_mode == 'readonly':
            raise RuntimeError('Cannot write data in readonly mode')
        if self._data_shape is None:
            self._data_shape = field.data.shape
        elif self.data_shape != field.data.shape:
            raise ValueError('Data shape incompatible with stored data')

        self._grid = field.grid
        self._field = field.copy()
        self.info['field_attributes'] = field.attributes_serialized 
#         self.info['grid'] = field.grid.state_serialized
            
    
    def end_writing(self) -> None:
        """ finalize the storage after writing """
        pass


    
class StorageTracker(TrackerBase):
    """ Tracker that stores data in special storage classes 
    
    Attributes:
        storage (:class:`~pde.storage.base.StorageBase`):
            The underlying storage class through which the data can be accessed 
    """

    @fill_in_docstring
    def __init__(self, storage, interval: IntervalData = 1):
        """
        Args:
            storage (:class:`~pde.storage.base.StorageBase`):
                Storage instance to which the data is written
            interval:
                {ARG_TRACKER_INTERVAL}
        """
        super().__init__(interval=interval)
        self.storage = storage
        
        
    def initialize(self, field: FieldBase, info: InfoDict = None) -> float:
        """ 
        Args:
            field (:class:`~pde.fields.FieldBase`):
                An example of the data that will be analyzed by the tracker
            info (dict):
                Extra information from the simulation        
                
        Returns:
            float: The first time the tracker needs to handle data
        """
        result = super().initialize(field, info)
        self.storage.start_writing(field, info)
        return result
        
        
    def handle(self, field: FieldBase, t: float) -> None:
        """ handle data supplied to this tracker
        
        Args:
            field (:class:`~pde.fields.FieldBase`):
                The current state of the simulation
            t (float): The associated time
        """
        self.storage.append(field.data, time=t)
    
    
    def finalize(self, info: InfoDict = None) -> None:
        """ finalize the tracker, supplying additional information

        Args:
            info (dict):
                Extra information from the simulation        
        """
        super().finalize(info)
        self.storage.end_writing()
    



'''
Spherically-symmetric grids in 2 and 3 dimensions. These are grids that only
discretize the radial direction, assuming symmetry with respect to all angles.

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
 
'''

from abc import ABCMeta
from typing import Tuple, Dict, Any, Union, Generator, TYPE_CHECKING

import numpy as np
from scipy import interpolate

from .base import GridBase, discretize_interval, _check_shape, DimensionError
from .cartesian import CartesianGrid
from ..tools.spherical import volume_from_radius
from ..tools.cache import cached_property
from ..tools.docstrings import fill_in_docstring
from ..tools.plotting import plot_on_axes



if TYPE_CHECKING:
    from .boundaries import Boundaries  # @UnusedImport



PI_4 = 4 * np.pi
PI_43 = 4 / 3 * np.pi



class SphericalGridBase(GridBase,  # lgtm [py/missing-equals]
                        metaclass=ABCMeta):
    r""" Base class for d-dimensional spherical grids with angular symmetry
    
    The angular symmetry implies that states only depend on the radial
    coordinate :math:`r`, which is discretized uniformly as
    
    
    .. math::
        r_i = R_\mathrm{inner} + \left(i + \frac12\right) \Delta r
        \quad \text{for} \quad i = 0, \ldots, N - 1
        \quad \text{with} \quad 
            \Delta r = \frac{R_\mathrm{outer} - R_\mathrm{inner}}{N}

    where :math:`R_\mathrm{outer}` is the outer radius of the grid and
    :math:`R_\mathrm{inner}` corresponds to a possible inner radius, which is
    zero by default. The radial direction is discretized by :math:`N` support
    points.    
    """

    periodic = [False]  # the radial axis is not periodic
    num_axes = 1        # the number of independent axes
    

    def __init__(self, radius: Union[float, Tuple[float, float]],
                 shape: Union[Tuple[int], int]):
        r""" 
        Args:
            radius (float or tuple of floats): 
                radius :math:`R_\mathrm{outer}` in case a simple float is given.
                If a tuple is supplied it is interpreted as the inner and outer
                radius, :math:`(R_\mathrm{inner}, R_\mathrm{outer})`.
            shape (tuple or int): A single number setting the number :math:`N`
                of support points along the radial coordinate
        """
        super().__init__()
        shape_list = _check_shape(shape)
        if not len(shape_list) == 1:
            raise ValueError('`shape` must be a single number, not '
                             f'{shape_list}')
        self._shape: Tuple[int] = (int(shape_list[0]),)
        
        try:
            r_inner, r_outer = radius  # type: ignore
        except TypeError:
            r_inner, r_outer = 0, float(radius)  # type: ignore
            
        if r_inner < 0:
            raise ValueError('Inner radius must be positive')
        if r_inner >= r_outer:
            raise ValueError('Outer radius must be larger than inner radius')
        
        # radial discretization
        rs, dr = discretize_interval(r_inner, r_outer, self.shape[0])
        
        self._axes_coords = (rs,)
        self._axes_bounds = ((r_inner, r_outer),)
        self._discretization = np.array((dr,))
        
        
    @property
    def state(self) -> Dict[str, Any]:
        """ state: the state of the grid """
        return {'radius': self.radius,
                'shape': self.shape}
        
        
    @property
    def has_hole(self) -> bool:
        """ returns whether the inner radius is larger than zero """
        return self.axes_bounds[0][0] > 0
        
        
    @classmethod
    def from_state(cls, state) -> "SphericalGridBase":
        """ create a field from a stored `state`.
        
        Args:
            state (dict):
                The state from which the grid is reconstructed.
        """
        state_copy = state.copy()
        obj = cls(radius=state_copy.pop('radius'),
                  shape=state_copy.pop('shape'))
        if state_copy:
            raise ValueError(f'State items {state_copy.keys()} were not used')
        return obj


    @property
    def radius(self) -> Union[float, Tuple[float, float]]:
        """ float: radius of the sphere """
        r_inner, r_outer = self.axes_bounds[0]
        if r_inner == 0:
            return r_outer
        else:
            return r_inner, r_outer
    

    @property
    def volume(self) -> float:
        """ float: total volume of the grid """
        r_inner, r_outer = self.axes_bounds[0]
        volume = volume_from_radius(r_outer, dim=self.dim)
        if r_inner > 0:
            volume -= volume_from_radius(r_inner, dim=self.dim)
        return volume


    @cached_property()
    def cell_volume_data(self) -> Tuple[np.ndarray]:
        """ tuple of :class:`numpy.ndarray`: the volumes of all cells """
        dr = self.discretization[0]
        rs = self.axes_coords[0]
        volumes_h = volume_from_radius(rs + 0.5 * dr, dim=self.dim)
        volumes_l = volume_from_radius(rs - 0.5 * dr, dim=self.dim)
        return ((volumes_h - volumes_l).reshape(self.shape[0]),)  # type: ignore
    

    def contains_point(self, point):
        """ check whether the point is contained in the grid
        
        Args:
            point (vector): Coordinates of the point
        """
        point = np.atleast_1d(point)
        if point.shape[-1] != self.dim:
            raise DimensionError('Dimension mismatch')
        r = np.linalg.norm(point, axis=-1)
        
        r_inner, r_outer = self.axes_bounds[0]
        return (r_inner <= r <= r_outer)

    
    def get_random_point(self, boundary_distance: float = 0,
                         cartesian: bool = True, avoid_center: bool = False):
        """ return a random point within the grid
        
        Note that these points will be uniformly distributed on the radial axis,
        which implies that they are not uniformly distributed in the volume.
        
        Args:
            boundary_distance (float): The minimal distance this point needs to
                have from all boundaries.
            cartesian (bool): Determines whether the point is returned in
                Cartesian coordinates or grid coordinates.
            avoid_center (bool): Determines whether the boundary distance
                should also be kept from the center, i.e., whether points close
                to the center are returned.
                
        Returns:
            :class:`numpy.ndarray`: The coordinates of the point        
        """
        # handle the boundary distance
        r_inner, r_outer = self.axes_bounds[0]
        r_min = r_inner
        if avoid_center:
            r_min += boundary_distance
        r_mag = r_outer - boundary_distance - r_min

        if r_mag <= 0:
            raise RuntimeError('Random points would be too close to boundary')

        # create random point
        r = np.array([r_mag * np.random.random() + r_min])
        if cartesian:
            return self.point_to_cartesian(r)
        else:
            return r


    def get_line_data(self, data, extract: str = 'auto') -> Dict[str, Any]:
        """ return a line cut along the radial axis
        
        Args:
            data (:class:`numpy.ndarray`):
                The values at the grid points
            extract (str):
                Determines which cut is done through the grid. This parameter is
                mainly supplied for a consistent interface and has no effect for
                polar grids.
            
        Returns:
            A dictionary with information about the line cut, which is 
            convenient for plotting.
        """
        if extract not in {'auto', 'r', 'radial'}:
            raise ValueError(f'Unknown extraction method `{extract}`')
        
        return {'data_x': self.axes_coords[0],
                'data_y': data,
                'extent_x': self.axes_bounds[0],
                'label_x': self.axes[0]}


    def get_image_data(self, data, performance_goal: str = 'speed',
                       fill_value: float = 0,
                       masked: bool = True) -> Dict[str, Any]:
        """ return a 2d-image of the data
        
        Args:
            data (:class:`numpy.ndarray`):
                The values at the grid points
            performance_goal (str):
                Determines the method chosen for interpolation. Possible options
                are `speed` and `quality`.
            fill_value (float):
                The value assigned to invalid positions (those inside the hole
                or outside the region).
            masked (bool):
                Whether a :class:`numpy.ma.MaskedArray` is returned for the data
                instead of the normal :class:`numpy.ndarray`.
            
        Returns:
            A dictionary with information about the image, which is  convenient
            for plotting.
        """
        _, r_outer = self.axes_bounds[0]
        r_data = self.axes_coords[0]
        
        if self.has_hole:
            num = int(np.ceil(r_outer / self.discretization[0]))
            x_positive, _ = discretize_interval(0, r_outer, num)
        else:
            x_positive = r_data
            
        x = np.r_[-x_positive[::-1], x_positive]
        xs, ys = np.meshgrid(x, x, indexing='ij')
        r_img = np.hypot(xs, ys)
        
        if performance_goal == 'speed':
            # interpolate over the new coordinates using linear interpolation
            f = interpolate.interp1d(r_data, data, copy=False,
                                     bounds_error=False,
                                     fill_value=fill_value,
                                     assume_sorted=True)
            data_int = f(r_img.flat).reshape(r_img.shape)
            
        elif performance_goal == 'quality':
            # interpolate over the new coordinates using radial base function
            f = interpolate.Rbf(r_data, data, function='cubic')
            data_int = f(r_img)
            
        else:
            raise ValueError(f'Performance goal `{performance_goal}` undefined')
        
        if masked:
            mask = (r_img < r_data[0]) | (r_data[-1] < r_img) 
            data_int = np.ma.masked_array(data_int, mask=mask)
        
        return {'data': data_int,
                'x': x, 'y': x,
                'extent': (-r_outer, r_outer, -r_outer, r_outer),
                'label_x': 'x', 'label_y': 'y'}       


    def iter_mirror_points(self, point, with_self: bool = False,
                           only_periodic: bool = True) -> Generator:
        """ generates all mirror points corresponding to `point`
        
        Args:
            point (:class:`numpy.ndarray`): the point within the grid
            with_self (bool): whether to include the point itself
            only_periodic (bool): whether to only mirror along periodic axes
        
        Returns:
            A generator yielding the coordinates that correspond to mirrors
        """
        point = np.asanyarray(point, dtype=np.double)
        if with_self:
            yield point


    def normalize_point(self, point, reduced_coords: bool = False):
        """ normalize coordinates, which is a no-op for spherical coordinates.
        
        Args:
            point (:class:`numpy.ndarray`): Coordinates of a single point
            reduced_coords (bool): Flag determining whether only the coordinates
                corresponding to axes in this grid are given
                 
        Returns:
            :class:`numpy.ndarray`: The input array
        """
        point = np.asarray(point, dtype=np.double)
        size = self.num_axes if reduced_coords else self.dim
        if point.size == 0:
            return np.zeros((0, size))
        if point.shape[-1] != size:
            raise DimensionError('Dimension mismatch: Array of shape '
                                 f'{point.shape} does not describe points of '
                                 f'dimension {size}.')
        return point

    
    def point_from_cartesian(self, points):
        """ convert points given in Cartesian coordinates to this grid
        
        Args:
            points (:class:`numpy.ndarray`):
                Points given in Cartesian coordinates.
                
        Returns:
            :class:`numpy.ndarray`: Points given in the coordinates of the grid
        """
        points = self.normalize_point(points)
        return np.linalg.norm(points, axis=-1, keepdims=True)


    def cell_to_point(self, cells, cartesian: bool = True):
        """ convert cell coordinates to real coordinates
        
        This function returns points restricted to the x-axis, i.e., the
        y-coordinate will be zero.
        
        Args:
            cells (:class:`numpy.ndarray`):
                Indices of the cells whose center coordinates are requested.
                This can be float values to indicate positions relative to the
                cell center.
            cartesian (bool):
                Determines whether the point is returned in Cartesian
                coordinates or grid coordinates.
                
        Returns:
            :class:`numpy.ndarray`: The center points of the respective cells
        """
        cells = np.atleast_1d(cells)
        # convert from cells indices to grid coordinates
        r_inner, _ = self.axes_bounds[0] 
        points = r_inner + (cells + 0.5) * self.discretization[0]
        if cartesian:
            return self.point_to_cartesian(points)
        else:
            return points


    def point_to_cell(self, points):
        """ Determine cell(s) corresponding to given point(s)

        This function respects periodic boundary conditions, but it does not
        throw an error when coordinates lie outside the bcs (for
        non-periodic axes).
        
        Args:
            points (:class:`numpy.ndarray`): Real coordinates
                
        Returns:
            :class:`numpy.ndarray`: The indices of the respective cells
        """
        # convert from grid coordinates to cells indices
        r_inner, _ = self.axes_bounds[0]
        r = self.point_from_cartesian(points)
        cells = (r - r_inner) / self.discretization[0]
        return cells.astype(np.int)

        
    def difference_vector_real(self, p1, p2):
        """ return the vector pointing from p1 to p2.
        
        In case of periodic boundary conditions, the shortest vector is returned
        
        Args:
            p1 (:class:`numpy.ndarray`): First point(s)
            p2 (:class:`numpy.ndarray`): Second point(s)
            
        Returns:
            :class:`numpy.ndarray`: The difference vectors between the points
                with periodic boundary conditions applied.
        """
        return np.atleast_1d(p2) - np.atleast_1d(p1)

    
    def polar_coordinates_real(self, origin=[0, 0, 0], ret_angle: bool = False):
        """ return spherical coordinates associated with the grid
        
        Args:
            origin (vector): Coordinates of the origin at which the polar
                coordinate system is anchored. Note that this must be of the 
                form `[0, 0, z_val]`, where only `z_val` can be chosen freely. 
            ret_angle (bool): Determines whether angles are returned alongside
                the distance. If `False` only the distance to the origin is 
                returned for each support point of the grid.
                If `True`, the distance and angles are returned. Note that in 
                the case of spherical grids, this angle is zero by convention.
        """
        origin = np.array(origin, dtype=np.double, ndmin=1)
        if not np.array_equal(origin, np.zeros(self.dim)):
            raise RuntimeError(f'Origin must be {str([0]*self.dim)}')
             
        # the distance to the origin is exactly the radial coordinate
        rs = self.axes_coords[0]
        if ret_angle:
            return rs, np.zeros_like(rs)
        else:
            return rs


    @fill_in_docstring
    def get_boundary_conditions(self, bc='natural', rank: int = 0) \
            -> "Boundaries":
        """ constructs boundary conditions from a flexible data format.
        
        Note that the inner boundary condition for grids without holes need not
        be specified.
        
        Args:
            bc (str or list or tuple or dict):
                The boundary conditions applied to the field.
                {ARG_BOUNDARIES}
            rank (int):
                The tensorial rank of the value associated with the boundary
                conditions.

        Raises:
            ValueError: If the data given in `bc` cannot be read
            PeriodicityError: If the boundaries are not compatible with the
                periodic axes of the grid. 
        """
        from .boundaries.local import BCBase, NeumannBC
        from .boundaries.axis import BoundaryPair
        from .boundaries import Boundaries  # @Reimport

        if isinstance(bc, Boundaries):
            return bc

        if self.has_hole:  # grid has holes => specify two boundary conditions
            if bc == 'natural':
                low = NeumannBC(self, 0, upper=False, rank=rank)
                high = NeumannBC(self, 0, upper=True, rank=rank)
                b_pair = BoundaryPair(low, high)
            else:
                b_pair = BoundaryPair.from_data(self, 0, bc, rank=rank)
            
        else:  # grid has no hole => need only one boundary condition
            b_inner = NeumannBC(self, 0, upper=False, rank=rank)
            if bc == 'natural':
                upper = NeumannBC(self, 0, upper=True, rank=rank)
                b_pair = BoundaryPair(b_inner, upper)
            else:
                try:
                    b_outer = BCBase.from_data(self, 0, upper=True, data=bc,
                                               rank=rank)
                except ValueError:
                    # this can happen when two boundary conditions have been
                    # supplied
                    b_pair = BoundaryPair.from_data(self, 0, bc, rank=rank)
                    if b_inner != b_pair.low:
                        raise ValueError(
                            f'Unsupported boundary format: `{bc}`. Note that '
                            'spherical symmetry implies vanishing derivatives '
                            'at the origin at r=0 for all fields. This '
                            'boundary condition need not be specified.')
                else:
                    b_pair = BoundaryPair(b_inner, b_outer)

        return Boundaries([b_pair])       

    
    def get_cartesian_grid(self, mode: str = 'valid', num: int = None) \
            -> CartesianGrid:
        """ return a Cartesian grid for this Cylindrical one 
        
        Args:
            mode (str):
                Determines how the grid is determined. Setting it to 'valid'
                only returns points that are fully resolved in the spherical
                grid, e.g., the sphere is circumscribed. Conversely, 'full'
                returns all data, so the sphere is inscribed. 
            num (int):
                Number of support points along each axis of the returned grid.
                
        Returns:
            :class:`pde.grids.cartesian.CartesianGrid`: The requested grid
        """
        # Pick the grid instance
        if mode == 'valid':
            if self.has_hole:
                self._logger.warn('The sphere has holes, so not all points are '
                                  'valid.')
            bounds = self.radius / np.sqrt(self.dim)
        elif mode == 'full':
            bounds = self.radius
        else:
            raise ValueError(f'Unsupported mode `{mode}`')
                
        # determine the grid points
        if num is None:
            num = 2 * round(bounds / self.discretization[0])
        grid_bounds = [(-bounds, bounds)] * self.dim
        return CartesianGrid(grid_bounds, num)
        


class PolarGrid(SphericalGridBase):
    r""" 2-dimensional polar grid assuming angular symmetry
    
    The angular symmetry implies that states only depend on the radial
    coordinate :math:`r`, which is discretized uniformly as
    
    .. math::
        r_i = R_\mathrm{inner} + \left(i + \frac12\right) \Delta r
        \quad \text{for} \quad i = 0, \ldots, N - 1
        \quad \text{with} \quad 
            \Delta r = \frac{R_\mathrm{outer} - R_\mathrm{inner}}{N}

    where :math:`R_\mathrm{outer}` is the outer radius of the grid and
    :math:`R_\mathrm{inner}` corresponds to a possible inner radius, which is
    zero by default. The radial direction is discretized by :math:`N` support
    points.    
    """
    
    dim = 2  # dimension of the described space
    axes = ['r']
    axes_symmetric = ['phi']
    coordinate_constraints = [0, 1]  # axes not described explicitly

        
    def point_to_cartesian(self, points):
        """ convert coordinates of a point to Cartesian coordinates
        
        This function returns points along the y-coordinate, i.e, the x
        coordinates will be zero.
                
        Returns:
            :class:`numpy.ndarray`: The Cartesian coordinates of the point
        """
        points = np.atleast_1d(points)
        if points.shape[-1] != 1:
            raise DimensionError(f'Dimension mismatch: Points {points} invalid')
        
        y = points[..., 0]
        x = np.zeros_like(y)
        return np.stack((x, y), axis=-1)


    @plot_on_axes()
    def plot(self, ax, **kwargs):
        r""" visualize the polar grid
        
        Args:
            ax (:class:`matplotlib.axes.Axes`):
                Figure axes to be used for plotting. If `None`, a new figure is
                created. This has no effect if a `reference` is supplied.
            \**kwargs: Extra arguments are passed on the to the matplotlib
                plotting routines, e.g., to set the color of the lines
        """
        from matplotlib import patches
        
        kwargs.setdefault('color', 'k')
        rb, = self.axes_bounds
        rmax = rb[1]

        for r in np.linspace(*rb, self.shape[0] + 1):
            if r == 0:
                ax.plot(0, 0, '.', **kwargs)
            else:
                c = patches.Circle((0, 0), r, fill=False, **kwargs)
                ax.add_artist(c)
        ax.set_xlim(-rmax, rmax)
        ax.set_xlabel('x')
        ax.set_ylim(-rmax, rmax)
        ax.set_ylabel('y')
        ax.set_aspect(1)
            
            
    
class SphericalGrid(SphericalGridBase):
    r""" 3-dimensional spherical grid assuming spherical symmetry
    
    The symmetry implies that states only depend on the radial coordinate 
    :math:`r`, which is discretized as follows:
    
    .. math::
        r_i = R_\mathrm{inner} + \left(i + \frac12\right) \Delta r
        \quad \text{for} \quad i = 0, \ldots, N - 1
        \quad \text{with} \quad 
            \Delta r = \frac{R_\mathrm{outer} - R_\mathrm{inner}}{N}

    where :math:`R_\mathrm{outer}` is the outer radius of the grid and
    :math:`R_\mathrm{inner}` corresponds to a possible inner radius, which is
    zero by default. The radial direction is discretized by :math:`N` support
    points.    
    """
    
    dim = 3  # dimension of the described space
    axes = ['r']
    axes_symmetric = ['theta', 'phi']
    coordinate_constraints = [0, 1, 2]  # axes not described explicitly
      
      
    def point_to_cartesian(self, points):
        """ convert coordinates of a point to Cartesian coordinates
        
        This function returns points along the z-coordinate, i.e, the x and y
        coordinates will be zero.
        
        Args:
            points (:class:`numpy.ndarray`):
                Points given in the coordinates of the grid
                
        Returns:
            :class:`numpy.ndarray`: The Cartesian coordinates of the point
        """
        points = np.atleast_1d(points)
        if points.shape[-1] != 1:
            raise DimensionError(f'Dimension mismatch: Points {points} invalid')
        z = points[..., 0]
        x = np.zeros_like(z)
        return np.stack((x, x, z), axis=-1)    



'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import numpy as np
import pytest

from .. import CylindricalGrid



def test_cylindrical_grid():
    """ test simple cylindrical grid """
    for periodic in [True, False]:
        grid = CylindricalGrid(4, (-1, 2), (8, 9), periodic_z=periodic)
        
        msg = str(grid)
        assert grid.dim == 3
        assert grid.numba_type == "f8[:, :]"
        assert grid.shape == (8, 9)
        assert grid.length == pytest.approx(3)
        assert grid.discretization[0] == pytest.approx(0.5)
        assert grid.discretization[1] == pytest.approx(1/3)
        np.testing.assert_array_equal(grid.discretization,
                                      np.array([0.5, 1/3]))
        assert grid.volume == pytest.approx(np.pi * 4**2 * 3)
        assert grid.volume == pytest.approx(grid.integrate(1))
        
        rs, zs = grid.axes_coords
        np.testing.assert_allclose(rs, np.linspace(0.25, 3.75, 8))
        np.testing.assert_allclose(zs, np.linspace(-1 + 1/6, 2 - 1/6, 9))
        
        # random points
        c = np.random.randint(8, size=(6, 2))
        c1 = grid.point_to_cell(grid.cell_to_point(c))
        np.testing.assert_almost_equal(c, c1, err_msg=msg)
        
        assert grid.contains_point(grid.get_random_point())
        assert grid.contains_point(grid.get_random_point(1.49))
        assert "laplace" in grid.operators
        


def test_cylindrical_to_cartesian():
    """ test conversion of cylindrical grid to Cartesian """
    from ...fields import ScalarField
    from .. import CartesianGrid
     
    expr_cyl = 'cos(z / 2) / (1 + r**2)'
    expr_cart = expr_cyl.replace('r**2', '(x**2 + y**2)')
     
    z_range = (-np.pi, 2*np.pi)
    grid_cyl = CylindricalGrid(10, z_range, (16, 33))
    pf_cyl = ScalarField.from_expression(grid_cyl, expression=expr_cyl)
    
    grid_cart = CartesianGrid([[-7, 7], [-6, 7], z_range], [16, 16, 16])
    pf_cart1 = pf_cyl.interpolate_to_grid(grid_cart)
    pf_cart2 = ScalarField.from_expression(grid_cart, expression=expr_cart) 
    np.testing.assert_allclose(pf_cart1.data, pf_cart2.data, atol=0.1)



def test_setting_domain_cylindrical():
    """ test various versions of settings bcs for cylindrical grids """
    grid = CylindricalGrid(1, [0, 1], [2, 2], periodic_z=False)
    grid.get_boundary_conditions('natural')
    grid.get_boundary_conditions(['no-flux', 'no-flux'])
    with pytest.raises(ValueError):
        grid.get_boundary_conditions(['no-flux'])
    with pytest.raises(ValueError):
        grid.get_boundary_conditions(['no-flux'] * 3)
    with pytest.raises(RuntimeError):
        grid.get_boundary_conditions(['no-flux', 'periodic'])

    grid = CylindricalGrid(1, [0, 1], [2, 2], periodic_z=True)
    grid.get_boundary_conditions('natural')
    grid.get_boundary_conditions(['no-flux', 'periodic'])
    with pytest.raises(RuntimeError):
        grid.get_boundary_conditions(['no-flux', 'no-flux'])
        
        

@pytest.mark.parametrize('periodic', [True, False])
def test_polar_conversion(periodic):
    """ test conversion to polar coordinates """
    grid = CylindricalGrid(1, [-1, 1], [5, 5], periodic_z=periodic)
    dists = grid.polar_coordinates_real([0, 0, 0])
    assert np.all(0.09 <= dists) 
    assert np.any(dists < 0.11)
    assert np.all(dists <= np.sqrt(2)) 
    assert np.any(dists > 0.8 * np.sqrt(2)) 
        



'''
Defines base classes


.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import functools
import operator
import logging
import json
from pathlib import Path 
from abc import ABCMeta, abstractmethod, abstractproperty
from typing import (Tuple, Callable, Optional, Union, Any, Dict, TypeVar,
                    TYPE_CHECKING)

import numpy as np
import numba as nb        
from scipy import interpolate, ndimage

from ..grids.base import (GridBase, discretize_interval, DimensionError,
                          DomainError)
from ..grids.cartesian import CartesianGridBase
from ..grids.boundaries.axes import BoundariesData
from ..tools.plotting import plot_on_axes, PlotReference
from ..tools.numba import jit, address_as_void_pointer
from ..tools.cache import cached_method
from ..tools.docstrings import fill_in_docstring


if TYPE_CHECKING:
    from .scalar import ScalarField  # @UnusedImport



ArrayLike = Union[np.ndarray, float]
OptionalArrayLike = Optional[ArrayLike]
TField = TypeVar('TField', bound='FieldBase')



class FieldBase(metaclass=ABCMeta):
    """ abstract base class for describing (discretized) fields
    
    Attributes:
        label (str):
            Name of the field
    """ 
    
    _subclasses: Dict[str, 'FieldBase'] = {}  # all classes inheriting from this
    readonly = False
    
    
    def __init__(self, grid: GridBase,
                 data: OptionalArrayLike = None,
                 label: Optional[str] = None):
        """ 
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            data (array, optional):
                Field values at the support points of the grid
            label (str, optional):
                Name of the field
        """
        self._grid = grid
        self._data: np.ndarray = data
        self.label = label
        self._logger = logging.getLogger(self.__class__.__name__)


    def __init_subclass__(cls, **kwargs):  # @NoSelf
        """ register all subclassess to reconstruct them later """
        super().__init_subclass__(**kwargs)
        cls._subclasses[cls.__name__] = cls


    @classmethod
    def from_state(cls, attributes: Dict[str, Any],
                   data: np.ndarray = None) -> "FieldBase":
        """ create a field from given state.
        
        Args:
            attributes (dict):
                The attributes that describe the current instance
            data (:class:`numpy.ndarray`, optional):
                Data values at the support points of the grid defining the field
        """
        # base class was chosen => select correct class from attributes
        class_name = attributes.pop('class')

        if class_name == cls.__name__:
            raise RuntimeError('Cannot reconstruct abstract class' 
                               f'`{class_name}`')
        
        # call possibly overwritten classmethod from subclass
        return cls._subclasses[class_name].from_state(attributes, data)

            
    @classmethod
    def from_file(cls, filename: str) -> "FieldBase":
        """ create field by reading file
        
        Args:
            filename (str): Path to the file being read
        """
        import h5py
        from .collection import FieldCollection
        
        with h5py.File(filename, "r") as fp:
            if 'class' in fp.attrs:
                # this should be a field collection
                assert json.loads(fp.attrs['class']) == 'FieldCollection'
                obj = FieldCollection._from_hdf_dataset(fp)
                
            elif len(fp) == 1:
                # a single field is stored in the data
                dataset = fp[list(fp.keys())[0]]  # retrieve only dataset
                obj = cls._from_hdf_dataset(dataset)  # type: ignore
                
            else:
                raise RuntimeError('Multiple data fields were found in the '
                                   'file but no FieldCollection is expected')
        return obj
                
                
    @classmethod
    def _from_hdf_dataset(cls, dataset) -> "FieldBase":
        """ construct a field by reading data from an hdf5 dataset """
        # copy attributes from hdf
        attributes = dict(dataset.attrs)
        
        # determine class
        class_name = json.loads(attributes.pop('class'))
        field_cls = cls._subclasses[class_name]
        
        # unserialize the attributes
        attributes = field_cls.unserialize_attributes(attributes)
        return field_cls.from_state(attributes, data=dataset)

    
    @property
    def grid(self) -> GridBase:
        """ GridBase: The grid on which the field is defined """
        return self._grid


    def to_file(self, filename: str, **kwargs):
        r""" store field in a file
        
        The extension of the filename determines what format is being used. If
        it ends in `.h5` or `.hdf`, the Hierarchical Data Format is used. The
        other supported format are images, where only the most typical formats
        are supported. 
        
        Args:
            filename (str):
                Path where the data is stored
            \**kwargs:
                Additional parameters may be supported for some formats 
        """
        extension = Path(filename).suffix.lower()
        
        if extension in {'.hdf', '.hdf5', '.he5', '.h5'}:
            import h5py
            with h5py.File(filename, "w") as fp:
                self._write_hdf_dataset(fp, **kwargs)
                
        elif extension in {'.png', '.jpg', '.jpeg', '.tif', '.pdf', '.svg'}:
            self._write_to_image(filename, **kwargs)
            
        else:
            raise ValueError('Do not know how to save data to file with '
                             f'extensions `{extension}`')


    def _write_hdf_dataset(self, hdf_path, key: str = 'data'):
        """ write data to a given hdf5 path `hdf_path` """
        # write the data
        dataset = hdf_path.create_dataset(key, data=self.data)

        # write attributes        
        for key, value in self.attributes_serialized.items():
            dataset.attrs[key] = value


    def _write_to_image(self, filename: str, **kwargs):
        """ write data to image 
        
        Args:
            filename (str): The path to the image that will be created
        """
        raise NotImplementedError(f'Cannot save {self.__class__.__name__} as '
                                  'an image')

    
    @abstractmethod
    def copy(self: TField, data: OptionalArrayLike = None, label: str = None) \
        -> TField: pass
            
         
    def assert_field_compatible(self, other: 'FieldBase',
                                accept_scalar: bool = False):
        """ checks whether `other` is compatible with the current field
        
        Args:
            other (FieldBase): Other field this is compared to
            accept_scalar (bool, optional): Determines whether it is acceptable
                that `other` is an instance of
                :class:`~pde.fields.ScalarField`.
        """
        from .scalar import ScalarField  # @Reimport
        
        # check whether they are the same class
        class_compatible = (self.__class__ == other.__class__ or
                            (accept_scalar and isinstance(other, ScalarField)))
        if not class_compatible:
            raise TypeError('Fields are incompatible')
        
        # check whether the associated grids are identical
        if not self.grid.compatible_with(other.grid):
            raise ValueError('Grids incompatible')
    
         
    @property
    def data(self) -> np.ndarray:
        """ :class:`numpy.ndarray`: discretized data at the support points """
        return self._data
    
    @data.setter
    def data(self, value):
        if self.readonly:
            raise RuntimeError(f'Cannot write to {self.__class__.__name__}')
        if isinstance(value, FieldBase):
            # copy data into current field
            self.assert_field_compatible(value, accept_scalar=True)
            self._data[:] = value.data
        else:
            self._data[:] = value
        
        
    @property
    def attributes(self) -> Dict[str, Any]:
        """ dict: describes the state of the instance (without the data) """
        return {'class': self.__class__.__name__,
                'grid': self.grid,
                'label': self.label}
    

    @property
    def attributes_serialized(self) -> Dict[str, str]:
        """ dict: serialized version of the attributes """
        results = {}
        for key, value in self.attributes.items():
            if key == 'grid':
                results[key] = value.state_serialized
            else:
                results[key] = json.dumps(value)
        return results
    
    
    @classmethod
    def unserialize_attributes(cls, attributes: Dict[str, str]) \
            -> Dict[str, Any]:
        """ unserializes the given attributes
        
        Args:
            attributes (dict):
                The serialized attributes
                
        Returns:
            dict: The unserialized attributes
        """
        # base class was chosen => select correct class from attributes
        class_name = json.loads(attributes['class'])

        if class_name == cls.__name__:
            raise RuntimeError('Cannot reconstruct abstract class' 
                               f'`{class_name}`')
        
        # call possibly overwritten classmethod from subclass
        return cls._subclasses[class_name].unserialize_attributes(attributes)


    @property
    def _data_flat(self):
        """ :class:`numpy.ndarray`: flat version of discretized data """
        # flatten the first dimension of the internal data
        return self._data.reshape(-1, *self.grid.shape)
    
    @_data_flat.setter
    def _data_flat(self, value):
        """ set the data from a value from a collection """
        if self.readonly:
            raise RuntimeError(f'Cannot write to {self.__class__.__name__}')
        # simply set the data -> this might need to be overwritten
        self._data = value
            
    
    def __eq__(self, other):
        """ test fields for equality, ignoring the label """
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.grid == other.grid and
                np.array_equal(self.data, other.data))
    

    def __neg__(self):
        """ return the negative of the current field """
        return self.copy(data=-self.data)


    def _binary_operation(self, other, op: Callable,
                          scalar_second: bool = True) -> "FieldBase":
        """ perform a binary operation between this field and `other`
        
        Args:
            other (number of FieldBase):
                The second term of the operator
            op (callable):
                A binary function calculating the result
            scalar_second (bool):
                Flag determining whether the second operator must be a scalar
                
        Returns:
            FieldBase: An field that contains the result of the operation. If
            `scalar_second == True`, the type of FieldBase is the same as `self`
        """
        if isinstance(other, FieldBase):
            # right operator is a field
            from .scalar import ScalarField  # @Reimport
            
            if scalar_second:
                # right operator must be a scalar 
                if not isinstance(other, ScalarField):
                    raise TypeError('Right operator must be a scalar field')
                self.grid.assert_grid_compatible(other.grid)
                result: FieldBase = self.copy(op(self.data, other.data))
                
            elif isinstance(self, ScalarField):
                # left operator is a scalar field (right can be tensor)
                self.grid.assert_grid_compatible(other.grid)
                result = other.copy(op(self.data, other.data))
                
            else:
                # left operator is tensor and right one might be anything
                self.assert_field_compatible(other, accept_scalar=True)
                result = self.copy(op(self.data, other.data))
                
        else:
            # the second operator is a number or a numpy array
            result = self.copy(op(self.data, other))
        return result        

    
    def _binary_operation_inplace(self, other, op_inplace: Callable,
                                  scalar_second: bool = True) -> "FieldBase":
        """ perform an in-place binary operation between this field and `other`
        
        Args:
            other (number of FieldBase):
                The second term of the operator
            op_inplace (callable):
                A binary function storing its result in the first argument
            scalar_second (bool):
                Flag determining whether the second operator must be a scalar.
                
        Returns:
            FieldBase: The field `self` with updated data
        """
        if self.readonly:
            raise RuntimeError(f'Cannot write to {self.__class__.__name__}')
        
        if isinstance(other, FieldBase):
            # right operator is a field
            from .scalar import ScalarField  # @Reimport
            
            if scalar_second:
                # right operator must be a scalar 
                if not isinstance(other, ScalarField):
                    raise TypeError('Right operator must be a scalar field')
                self.grid.assert_grid_compatible(other.grid)
            else:
                # left operator is tensor and right one might be anything
                self.assert_field_compatible(other, accept_scalar=True)
                
            op_inplace(self.data, other.data)
                
        else:
            # the second operator is a number or a numpy array
            op_inplace(self.data, other)
            
        return self        

    
    def __add__(self, other) -> "FieldBase":
        """ add two fields """
        return self._binary_operation(other, operator.add, scalar_second=False)
    __radd__ = __add__


    def __iadd__(self, other) -> "FieldBase":
        """ add `other` to the current field """
        return self._binary_operation_inplace(other, operator.iadd,
                                              scalar_second=False)
    
    
    def __sub__(self, other) -> "FieldBase":
        """ subtract two fields """
        return self._binary_operation(other, operator.sub, scalar_second=False)

    
    def __rsub__(self, other) -> "FieldBase":
        """ subtract two fields """
        return self._binary_operation(other, lambda x, y: y - x,
                                      scalar_second=False)


    def __isub__(self, other) -> "FieldBase":
        """ add `other` to the current field """
        return self._binary_operation_inplace(other, operator.isub,
                                              scalar_second=False)
    
    
    def __mul__(self, other) -> "FieldBase":
        """ multiply field by value """
        return self._binary_operation(other, operator.mul, scalar_second=False)
    __rmul__ = __mul__
    
       
    def __imul__(self, other) -> "FieldBase":
        """ multiply field by value """
        return self._binary_operation_inplace(other, operator.imul,
                                              scalar_second=False)

    
    def __truediv__(self, other) -> "FieldBase":
        """ divide field by value """
        return self._binary_operation(other, operator.truediv,
                                      scalar_second=True)
    
    
    def __itruediv__(self, other) -> "FieldBase":
        """ divide field by value """
        return self._binary_operation_inplace(other, operator.itruediv,
                                              scalar_second=True)
       
       
    def __pow__(self, exponent: float) -> "FieldBase":
        """ raise data of the field to a certain power """
        if not np.isscalar(exponent):
            raise NotImplementedError('Only scalar exponents are supported')
        return self.copy(data=self.data ** exponent)
    
    
    def __ipow__(self, exponent: float) -> "FieldBase":
        """ raise data of the field to a certain power in-place """
        if self.readonly:
            raise RuntimeError(f'Cannot write to {self.__class__.__name__}')
        
        if not np.isscalar(exponent):
            raise NotImplementedError('Only scalar exponents are supported')
        self.data **= exponent
        return self
    

    @abstractmethod
    def get_line_data(self, scalar: str = 'auto', extract: str = 'auto'): pass

    @abstractmethod
    def get_image_data(self): pass
    
    @abstractmethod
    def plot(self, *args, **kwargs): pass
    


TDataField = TypeVar('TDataField', bound='DataFieldBase')


class DataFieldBase(FieldBase, metaclass=ABCMeta):
    """ abstract base class for describing fields of single entities
    
    Attributes:
        grid (:class:`~pde.grids.GridBase`):
            The underlying grid defining the discretization
        data (:class:`numpy.ndarray`):
            Data values at the support points of the grid
        shape (tuple):
            Shape of the `data` field
        label (str):
            Name of the field
    """ 
         
    rank: int  # the rank of the tensor field
    _allocate_memory = True  # determines whether the instances allocated memory
             

    def __init__(self, grid: GridBase,
                 data: OptionalArrayLike = None,
                 label: Optional[str] = None):
        """ 
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            data (array, optional):
                Field values at the support points of the grid. The data is
                copied from the supplied array.
            label (str, optional):
                Name of the field
        """
        # determine data shape
        shape = (grid.dim,) * self.rank + grid.shape
        
        if self._allocate_memory:
            # class manages its own data, which therefore needs to be allocated
            if data is None:
                data = np.zeros(shape, dtype=np.double)
            elif isinstance(data, DataFieldBase):
                # we need to make a copy to make sure the data is writeable 
                data = np.array(np.broadcast_to(data.data, shape),
                                dtype=np.double, copy=True)
            else:
                # we need to make a copy to make sure the data is writeable 
                data = np.array(np.broadcast_to(data, shape), dtype=np.double,
                                copy=True)
                
        elif data is not None:
            # class does not manage its own data
            raise ValueError(f"{self.__class__.__name__} does not support data "
                              "assignment.")
            
        super().__init__(grid, data=data, label=label)
            
                     
    def __repr__(self):
        """ return instance as string """
        class_name = self.__class__.__name__
        result = f'{class_name}(grid={self.grid!r}, data={self.data}'
        if self.label:
            result += f', label="{self.label}"'
        return result + ')'

    
    def __str__(self):
        """ return instance as string """
        result = f'{self.__class__.__name__}(grid={self.grid}, ' \
                 f'data=Array{self.data.shape}'
        if self.label:
            result += f', label="{self.label}"'
        return result + ')'


    @classmethod
    def random_uniform(cls, grid: GridBase, vmin: float = 0, vmax: float = 1,
                       label: Optional[str] = None, seed: Optional[int] = None):
        """ create field with uniform distributed random values
        
        These values are uncorrelated in space.
        
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            vmin (float):
                Smallest possible random value
            vmax (float):
                Largest random value
            label (str, optional):
                Name of the field
            seed (int, optional):
                Seed of the random number generator. If `None`, the current
                state is not changed.
        """
        shape = (grid.dim,) * cls.rank + grid.shape
        if seed is not None:
            np.random.seed(seed)
        data = np.random.uniform(vmin, vmax, shape)
        return cls(grid, data, label=label)
    
    
    @classmethod
    def random_normal(cls, grid: GridBase, mean: float = 0, std: float = 1,
                      scaling: str = 'physical', label: Optional[str] = None,
                      seed: Optional[int] = None):
        """ create field with normal distributed random values
        
        These values are uncorrelated in space.
        
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            mean (float):
                Mean of the Gaussian distribution
            std (float):
                Standard deviation of the Gaussian distribution
            scaling (str):
                Determines how the values are scaled. Possible choices are
                'none' (values are drawn from a normal distribution with
                given mean and standard deviation) or 'physical' (the variance
                of the random number is scaled by the inverse volume of the grid
                cell; this is useful for physical quantities, which vary less in
                larger volumes).            
            label (str, optional):
                Name of the field
            seed (int, optional):
                Seed of the random number generator. If `None`, the current
                state is not changed.
        """
        if seed is not None:
            np.random.seed(seed)
        
        if scaling == 'none':
            noise_scale = std
        elif scaling == 'physical':
            noise_scale = std / np.sqrt(grid.cell_volumes)
        else:
            raise ValueError(f'Unknown noise scaling {scaling}')
            
        shape = (grid.dim,) * cls.rank + grid.shape
        data = mean + noise_scale * np.random.randn(*shape)
        return cls(grid, data, label=label)

    
    @classmethod
    def random_harmonic(cls, grid: GridBase,
                        modes: int = 3,
                        harmonic=np.cos,
                        axis_combination=np.multiply,
                        label: Optional[str] = None,
                        seed: Optional[int] = None):
        r""" create a random field build from harmonics
        
        The resulting fields will be highly correlated in space and can thus
        serve for testing differential operators. 
        
        With the default settings, the resulting field :math:`c_i(\mathbf{x})`
        is given by
        
        .. math::
            c_i(\mathbf{x}) = \prod_{\alpha=1}^N \sum_{j=1}^M a_{ij\alpha}
                \cos\left(\frac{2 \pi x_\alpha}{j L_\alpha}\right) \;,
            
        where :math:`N` is the number of spatial dimensions, each with length 
        :math:`L_\alpha`, :math:`M` is the number of modes given by `modes`, and
        :math:`a_{ij\alpha}` are random amplitudes, chosen from a uniform
        distribution over the interval [0, 1].
        
        Note that the product could be replaced by a sum when
        `axis_combination = numpy.add` and the :math:`\cos()` could be any other
        function given by the parameter `harmonic`.
        
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            modes (int):
                Number :math:`M` of harmonic modes
            harmonic (callable):
                Determines which harmonic function is used. Typical values are
                :func:`numpy.sin` and :func:`numpy.cos`, which basically relate
                to different boundary conditions applied at the grid boundaries.
            axis_combination (callable):
                Determines how values from different axis are combined. Typical 
                choices are :func:`numpy.multiply` and :func:`numpy.add`
                resulting in products and sums of the values along axes,
                respectively.
            label (str, optional):
                Name of the field
            seed (int, optional):
                Seed of the random number generator. If `None`, the current
                state is not changed.
        """
        tensor_shape = (grid.dim,) * cls.rank
        if seed is not None:
            np.random.seed(seed)
    
        data = np.empty(tensor_shape + grid.shape)
        # determine random field for each component
        for index in np.ndindex(*tensor_shape):
            data_axis = []
            # random harmonic function along each axis
            for i in range(len(grid.axes)):
                # choose wave vectors
                ampl = np.random.random(modes)  # amplitudes
                x = discretize_interval(0, 2*np.pi, grid.shape[i])[0]
                data_axis.append(sum(a * harmonic(n * x)
                                     for n, a in enumerate(ampl, 1)))
            # full dataset is product of values along axes
            data[index] = functools.reduce(axis_combination.outer, data_axis)
            
        return cls(grid, data, label=label)


    @classmethod
    def random_colored(cls, grid: GridBase,
                       exponent: float = 0,
                       scale: float = 1,
                       label: Optional[str] = None, 
                       seed: Optional[int] = None):
        r""" create a field of random values with colored noise
                
        The spatially correlated values obey
        
        .. math::
            \langle c_i(\boldsymbol k) c_j(\boldsymbol k’) \rangle =
                \Gamma^2 |\boldsymbol k|^\nu \delta_{ij}
                \delta(\boldsymbol k - \boldsymbol k’)
                
        in spectral space. The special case :math:`\nu = 0` corresponds to white
        noise. Note that the components of vector or tensor fields are
        uncorrelated.
         
        Args:
            grid (:class:`~pde.grids.GridBase`):
                Grid defining the space on which this field is defined
            exponent (float):
                Exponent :math:`\nu` of the power spectrum
            scale (float):
                Scaling factor :math:`\Gamma` determining noise strength
            label (str, optional):
                Name of the field
            seed (int, optional):
                Seed of the random number generator. If `None`, the curren
                 state is not changed.
        """
        if seed is not None:
            np.random.seed(seed)
        
        # create function making colored noise    
        from pde.tools.spectral import make_colored_noise
        make_noise = make_colored_noise(grid.shape, dx=grid.discretization,
                                        exponent=exponent, scale=scale)
        
        # create random fields for each tensor component
        tensor_shape = (grid.dim,) * cls.rank
        data = np.empty(tensor_shape + grid.shape)
        # determine random field for each component
        for index in np.ndindex(*tensor_shape):
            data[index] = make_noise()
         
        return cls(grid, data, label=label)
    

    @classmethod
    def from_state(cls, attributes: Dict[str, Any],
                   data: np.ndarray = None) -> "DataFieldBase":
        """ create a field from given state.
        
        Args:
            attributes (dict):
                The attributes that describe the current instance
            data (:class:`numpy.ndarray`, optional):
                Data values at the support points of the grid defining the field
        """
        if 'class' in attributes:
            class_name = attributes.pop('class')
            assert class_name == cls.__name__
            
        # create the instance from the attributes
        return cls(attributes.pop('grid'), data=data, **attributes)
        
        
    def copy(self: TDataField,
             data: OptionalArrayLike = None,
             label: str = None) -> TDataField:
        """ return a copy of the data, but not of the grid
        
        Args:
            data (:class:`numpy.ndarray`, optional):
                Data values at the support points of the grid that define the
                field.
            label (str, optional):
                Name of the copied field
        """
        if label is None:
            label = self.label
        if data is None:
            data = self.data
        # the actual data will be copied in our __init__ method
        return self.__class__(self.grid, data=data, label=label)
    
    
    @property
    def data_shape(self) -> Tuple[int, ...]:
        """ tuple: the shape of the data at each grid point """
        return (self.grid.dim,) * self.rank
    
    
    @classmethod
    def unserialize_attributes(cls, attributes: Dict[str, str]) \
            -> Dict[str, Any]:
        """ unserializes the given attributes
        
        Args:
            attributes (dict):
                The serialized attributes
                
        Returns:
            dict: The unserialized attributes
        """
        results = {}
        for key, value in attributes.items():
            if key == 'grid':
                results[key] = GridBase.from_state(value)
            else:
                results[key] = json.loads(value)
        return results
    
        
    def _write_to_image(self, filename: str, **kwargs):
        r""" write data to image 
        
        Args:
            filename (str):
                The path to the image that will be created
            \**kwargs:
                Additional keyword arguments that affect the image. For
                instance, some fields support a `scalar` argument that
                determines how they are converted to a scalar. Non-Cartesian
                grids might support a `performance_goal` argument to influence
                how an image is created from the raw data. Finally, the
                remaining arguments are are passed to
                :func:`matplotlib.pyplot.imsave` to affect the appearance.
        """
        import matplotlib.pyplot as plt
        # obtain image data
        get_image_args = {}
        for key in ['performance_goal', 'scalar']:
            if key in kwargs:
                get_image_args[key] = kwargs.pop(key)
        img = self.get_image_data(**get_image_args)
        
        kwargs.setdefault('cmap', 'gray')
        plt.imsave(filename, img['data'], origin='lower', **kwargs)
    

    def _make_interpolator_scipy(self, method: str = 'linear',
                                 fill: float = None, **kwargs) -> Callable:
        r""" returns a function that can be used to interpolate values.
        
        This uses scipy.interpolate.RegularGridInterpolator and the
        interpolation method can thus be chosen when calling the returned
        function using the `method` keyword argument. Keyword arguments are
        directly forwarded to the constructor of `RegularGridInterpolator`.
        
        Note that this interpolator does not respect periodic boundary
        conditions, yet.
        
        Args:
            method (str):
                The method used for interpolation. If 'linear' or 'nearest'
                :class:`scipy.interpolate.RegularGridInterpolator` is used with
                the specified method. If 'rbf', :class:`scipy.interpolate.Rbf`
                is used for radial basis function interpolation.
            fill (float, optional):
                Determines how values out of bounds are handled. If `None`, a
                `ValueError` is raised when out-of-bounds points are requested.
                Otherwise, the given value is returned.
            \**kwargs: All keyword arguments are forwarded to
                :class:`scipy.interpolate.RegularGridInterpolator`
                
        Returns:
            A function which returns interpolated values when called with
            arbitrary positions within the space of the grid. 
        """
        coords_src = self.grid.axes_coords
        grid_dim = len(self.grid.axes)
        
        if self.rank == 0:
            # scalar field => data layout is already usable
            data = self.data
            revert_shape = False
        else:
            # spatial dimensions need to come first => move data to last axis
            assert self.data.shape[:-grid_dim] == self.data_shape
            data_flat = self._data_flat
            data_flat = np.moveaxis(data_flat, 0, -1)
            new_shape = self.grid.shape + (-1,)
            data = data_flat.reshape(new_shape)
            assert data.shape[-1] == self.grid.dim ** self.rank
            revert_shape = True
            
        # set the fill behavior
        if fill is None:
            kwargs['bounds_error'] = True
        else:
            kwargs['bounds_error'] = False
            kwargs['fill_value'] = fill
            
        # prepare the interpolator
        intp = interpolate.RegularGridInterpolator(coords_src, data, **kwargs)
        
        # determine under which conditions the axes can be squeezed
        if grid_dim == 1:
            scalar_dim = 0
        else:
            scalar_dim = 1
            
        # introduce wrapper function to process arrays
        def interpolator(point, **kwargs):
            """ return the interpolated value at the position `point` """
            point = np.atleast_1d(point)
            # apply periodic boundary conditions to grid point
            point = self.grid.normalize_point(point, reduced_coords=True)
            out = intp(point, **kwargs)
            if point.ndim == scalar_dim or point.ndim == point.size == 1:
                out = out[0]
            if revert_shape:
                # revert the shuffling of spatial and local axes
                out = np.moveaxis(out, point.ndim - 1, 0) 
                out = out.reshape(self.data_shape + point.shape[:-1])
            
            return out
            
        return interpolator


    @fill_in_docstring
    def _make_interpolator_compiled(self, bc: BoundariesData = 'natural',
                                    fill: float = None) -> Callable:
        """ return a compiled interpolator
        
        This interpolator respects boundary conditions and can thus interpolate
        values in the whole grid volume. However, close to corners, the
        interpolation might not be optimal, in particular for periodic grids.
        
        Args:
            bc:
                The boundary conditions applied to the field. {ARG_BOUNDARIES}
            fill (float, optional):
                Determines how values out of bounds are handled. If `None`, a
                `ValueError` is raised when out-of-bounds points are requested.
                Otherwise, the given value is returned.
                
        Returns:
            A function which returns interpolated values when called with
            arbitrary positions within the space of the grid. 
        """
        grid = self.grid
        grid_dim = len(grid.axes)
        data_shape = self.data_shape

        # create an interpolator for a single point
        if fill is not None:
            if self.rank == 0:
                fill = float(fill)
            else:
                fill = np.broadcast_to(fill, self.data_shape).astype(float)
        interpolate_single = \
            grid.make_interpolator_compiled(bc=bc, rank=self.rank, fill=fill)
        
        # extract information about the data field
        data_addr = self.data.ctypes.data
        shape, dtype = self.data.shape, self.data.dtype
        
        @jit
        def interpolator(point: np.ndarray, data: np.ndarray = None) \
                -> np.ndarray:
            """ return the interpolated value at the position `point`
             
            Args:
                point (:class:`numpy.ndarray`):
                    The list of points. This point coordinates should be given
                    along the last axis, i.e., the shape should be `(..., dim)`.
                data (:class:`numpy.ndarray`, optional):
                    The discretized field values. If omitted, the data of the 
                    current field is used, which should be the default. However,
                    this option can be useful to interpolate other fields
                    defined on the same grid without recreating the interpolator 
                     
            Returns:
                :class:`numpy.ndarray`: The interpolated values at the points
            """
            # check input
            point = np.atleast_1d(point)
            if point.shape[-1] != grid_dim:
                raise DimensionError('Dimension of the interpolation point '
                                     'does not match grid dimension')
            point_shape = point.shape[:-1]
            
            # reconstruct data field from memory address
            if data is None:
                data = nb.carray(address_as_void_pointer(data_addr), shape,
                                 dtype)
             
            # interpolate at every point
            out = np.empty(data_shape + point_shape)
            for idx in np.ndindex(point_shape):
                out[(...,) + idx] = interpolate_single(data, point[idx])
             
            return out

        return interpolator  # type: ignore


    @cached_method()
    def make_interpolator(self, method: str = 'numba',
                          fill: float = None,
                          **kwargs) -> Callable:
        r""" returns a function that can be used to interpolate values.
        
        Args:
            method (str):
                Determines the method being used for interpolation.
                Possible values are:
                
                * `scipy_nearest`: Use scipy to interpolate to nearest neighbors
                * `scipy_linear`: Linear interpolation using scipy
                * `numba`: Linear interpolation using numba (default)

            fill (float, optional):
                Determines how values out of bounds are handled. If `None`, a
                `ValueError` is raised when out-of-bounds points are requested.
                Otherwise, the given value is returned.
            \**kwargs:
                Additional keyword arguments are passed to the individual
                interpolator methods and can be used to further affect the
                behavior.
                  
        The scipy implementations use scipy.interpolate.RegularGridInterpolator
        and thus do not respect boundary conditions. Additional keyword
        arguments are directly forwarded to the constructor of
        `RegularGridInterpolator`.
        
        The numba implementation respect boundary conditions, which can be set
        using the `bc` keywords argument. Supported values are the same as for
        the operators, e.g., the Laplacian. If no boundary conditions are
        specified,  natural boundary conditions are assumed, which are periodic
        conditions for periodic axes and Neumann conditions otherwise.
        
        Returns:
            A function which returns interpolated values when called with
            arbitrary positions within the space of the grid. 
        """
        if method.startswith('scipy'):
            return self._make_interpolator_scipy(method=method[6:], fill=fill,
                                                 **kwargs)
        elif method == 'numba':
            return self._make_interpolator_compiled(fill=fill, **kwargs)
        else:
            raise ValueError(f'Unknown interpolation method `{method}`')
            

    def interpolate(self, point, method: str = 'numba', fill: float = None,
                    **kwargs):
        r""" interpolate the field to points between support points
        
        Args:
            point (:class:`numpy.ndarray`):
                The points at which the values should be obtained. This is given
                in grid coordinates.
            method (str):
                Determines the method being used for interpolation.
            fill (float, optional):
                Determines how values out of bounds are handled. If `None`, a
                `ValueError` is raised when out-of-bounds points are requested.
                Otherwise, the given value is returned.
            \**kwargs:
                Additional keyword arguments are forwarded to the method
                :meth:`DataFieldBase.make_interpolator`.
                
        Returns:
            :class:`numpy.ndarray`: the values of the field
        """
        point = np.asarray(point)
        return self.make_interpolator(method=method, fill=fill, **kwargs)(point)


    def interpolate_to_grid(self: TDataField, grid: GridBase,
                            method: str = 'numba',
                            fill: float = None,
                            label: Optional[str] = None) -> TDataField:
        """ interpolate the data of this field to another grid.
        
        Args:
            grid (:class:`~pde.grids.GridBase`):
                The grid of the new field onto which the current field is
                interpolated.
            method (str):
                Specifies interpolation method, e.g., 'numba', 'scipy_linear',
                'scipy_nearest' .
            fill (float, optional):
                Determines how values out of bounds are handled. If `None`, a
                `ValueError` is raised when out-of-bounds points are requested.
                Otherwise, the given value is returned.
            label (str, optional):
                Name of the returned field
            
        Returns:
            Field of the same rank as the current one.
        """
        if self.grid.dim != grid.dim:
            raise DimensionError(f'Grid dimensions are incompatible '
                                 f'({self.grid.dim:d} != {grid.dim:d})')
            
        # determine the points at which data needs to be calculated
        if isinstance(grid, CartesianGridBase):
            # convert to a Cartesian grid
            points = self.grid.point_from_cartesian(grid.cell_coords)
            
        elif self.grid.__class__ is grid.__class__:
            # convert within the same grid class
            points = grid.cell_coords
            
        else:
            # this type of interpolation is not supported
            raise NotImplementedError('Cannot convert '
                                      f'{self.grid.__class__.__name__} to '
                                      f'{grid.__class__.__name__}')
        
        # interpolate the data to the grid
        data = self.interpolate(points, method, fill)
        return self.__class__(grid, data, label=label)
    
    
    def add_interpolated(self, point, amount):
        """ adds an (integrated) value to the field at an interpolated position
        
        Args:
            point (:class:`numpy.ndarray`):
                The point inside the grid where the value is added. This is
                given in grid coordinates.
            amount (float or :class:`numpy.ndarray`):
                The amount that will be added to the field. The value describes
                an integrated quantity (given by the field value times the
                discretization volume). This is important for consistency with
                different discretizations and in particular grids with
                non-uniform discretizations.
        """
        point = np.atleast_1d(point)
        amount = np.broadcast_to(amount, self.data_shape)
        grid = self.grid
        grid_dim = len(grid.axes)
    
        if point.size != grid_dim or point.ndim != 1:
            raise DimensionError(f'Dimension mismatch for point {point}')
        
        point = grid.normalize_point(point, reduced_coords=True)
    
        low = np.array(grid.axes_bounds)[:, 0]
        c_l, d_l = np.divmod((point - low) / grid.discretization - 0.5, 1.)
        c_l = c_l.astype(np.int)
        w_l = 1 - d_l  # weights of the low point
        w_h = d_l      # weights of the high point
    
        # determine the total weight in first iteration
        total_weight = 0
        cells = []
        for i in np.ndindex(*((2,) * grid_dim)):
            coords = np.choose(i, [c_l, c_l + 1])
            if np.all(coords >= 0) and np.all(coords < grid.shape):
                weight = np.prod(np.choose(i, [w_l, w_h]))
                total_weight += weight
                cells.append((tuple(coords), weight))
                
        if total_weight == 0:
            raise DomainError('Point lies outside grid')
    
        # alter each point in second iteration
        for coords, weight in cells:
            chng = weight * amount / (total_weight * grid.cell_volumes[coords])
            self.data[(Ellipsis,) + coords] += chng                                                  
        
        
    @fill_in_docstring
    def get_boundary_values(self, axis: int,
                            upper: bool,
                            bc: BoundariesData = 'natural') -> np.ndarray:
        """ get the field values directly on the specified boundary 
        
        Args:
            axis (int):
                The axis perpendicular to the boundary
            upper (bool):
                Whether the boundary is at the upper side of the axis 
            bc:
                The boundary conditions applied to the field. {ARG_BOUNDARIES}
            
        Returns:
            :class:`~numpy.ndarray`: The discretized values on the boundary
        """
        interpolator = self.make_interpolator(bc=bc)
        points = self.grid._boundary_coordinates(axis, upper)
        return interpolator(points)
        
        
    @fill_in_docstring
    def make_get_boundary_values(self, axis: int,
                                 upper: bool,
                                 bc: BoundariesData = 'natural') -> np.ndarray:
        """ make a function calculating field values on the specified boundary 
        
        Args:
            axis (int):
                The axis perpendicular to the boundary
            upper (bool):
                Whether the boundary is at the upper side of the axis 
            bc:
                The boundary conditions applied to the field. {ARG_BOUNDARIES}
            
        Returns:
            callable: A function returning the values on the boundary. The
            function has the signature `(data=None, out=None)`, which allows
            specifying an input and an output :class:`~numpy.ndarray`. If `data`
            is omitted, the data of the current field is used. The resulting
            interpolation is written to `out` if it is present. Otherwise, a new
            array is created.
        """
        interpolator = self.make_interpolator(bc=bc)
        points = self.grid._boundary_coordinates(axis, upper)
        
        # TODO: use jit_allocated_out with pre-calculated shape
        
        @jit
        def get_boundary_values(data: np.ndarray = None,
                                out: np.ndarray = None):
            """ interpolate the field at the boundary
            
            Args:
                data (:class:`~numpy.ndarray`, optional):
                    The data values that are used for interpolation. The data of
                    the current field is used if `data = None`.
                out (:class:`~numpy.ndarray`, optional):
                    The array into which the interpolated results are written. A
                    new array is created if `out = None`.
                    
            Returns:
                :class:`numpy.ndarray`: The interpolated values on the boundary.
            """
            res = interpolator(points, data)
            if out is None:
                return res
            else:
                # the following just copies the data from res to out. It is a
                # workaround for a bug in numba existing up to at least ver 0.49
                out[...] = res[()]
                return out
            
        return get_boundary_values
        
        
    def apply(self: TDataField, func: Callable,
              out: Optional[TDataField] = None,
              label: str = None) -> TDataField:
        """ applies a function to the data and returns it as a field
        
        Args:
            func (callable):
                The (vectorized) function being applied to the data.
            out (FieldBase, optional):
                Optional field into which the data is written
            label (str, optional):
                Name of the returned field

        Returns:
            Field with new data. This is stored at `out` if given. 
        """
        if out is None:
            return self.copy(data=func(self.data), label=label)
        else:
            self.assert_field_compatible(out)
            out.data[:] = func(self.data)
            if label:
                out.label = label
            return out
        
        
    @abstractproperty
    def integral(self) -> Union[np.ndarray, float]: pass 
    
    @abstractmethod
    def to_scalar(self, scalar: str = 'auto',
                  label: Optional[str] = None) -> "ScalarField": pass
                  
                  
    @property
    def average(self) -> Union[np.ndarray, float]:
        """ determine the average of data
        
        This is calculated by integrating each component of the field over space
        and dividing by the grid volume
        """
        return self.integral / self.grid.volume


    @property    
    def fluctuations(self):
        """ :class:`numpy.ndarray`: fluctuations over the entire space.
        
        The fluctuations are defined as the standard deviation of the data
        scaled by the cell volume. This definition makes the fluctuations
        independent of the discretization. It corresponds to the physical
        scaling available in the :func:`~DataFieldBase.random_normal`.
        
        Returns:
            :class:`numpy.ndarray`: A tensor with the same rank of the field,
            specifying the fluctuations of each component of the tensor field
            individually. Consequently, a simple scalar is returned for a
            :class:`~pde.fields.scalar.ScalarField`.
        """
        scaled_data = self.data * np.sqrt(self.grid.cell_volumes)
        axes = tuple(range(self.rank, self.data.ndim))
        return np.std(scaled_data, axis=axes)
            
    
    @property
    def magnitude(self) -> float:
        """ float: determine the magnitude of the field.
        
        This is calculated by getting a scalar field using the default arguments
        of the :func:`to_scalar` method and averaging the result over the whole
        grid.
        """
        if self.rank == 0:
            return float(self.average)
        elif self.rank > 0:
            return self.to_scalar().average
        else:
            raise NotImplementedError('Magnitude cannot be determined for '
                                      'field ' + self.__class__.__name__)
            
            
    def smooth(self: TDataField, sigma: Optional[float] = 1,
               out: Optional[TDataField] = None,
               label: str = None) -> TDataField:
        """ applies Gaussian smoothing with the given standard deviation

        This function respects periodic boundary conditions of the underlying
        grid, using reflection when no periodicity is specified.
        
        sigma (float, optional):
            Gives the standard deviation of the smoothing in real length units
            (default: 1)
        out (FieldBase, optional):
            Optional field into which the smoothed data is stored. Setting this
            to the input field enables in-place smoothing.
        label (str, optional):
            Name of the returned field

        Returns:
            Field with smoothed data. This is stored at `out` if given.             
        """
        # allocate memory for storing output
        data_in = self._data
        if out is None:
            out = self.__class__(self.grid, label=self.label)
        else:
            self.assert_field_compatible(out)
        
        # apply Gaussian smoothing for each axis
        data_out = out._data
        for axis in range(-len(self.grid.axes), 0):
            sigma_dx = sigma / self.grid.discretization[axis]
            mode = 'wrap' if self.grid.periodic[axis] else 'reflect'
            ndimage.gaussian_filter1d(data_in, sigma=sigma_dx, axis=axis,
                                      output=data_out, mode=mode)
            data_in = data_out
            
        # return the data in the correct field class
        if label:
            out.label = label
        return out


    def get_line_data(self, scalar: str = 'auto', extract: str = 'auto') \
            -> Dict[str, Any]:
        """ return data for a line plot of the field
        
        Args:
            scalar (str or int):
                The method for extracting scalars as described in
                :meth:`DataFieldBase.to_scalar`.
            extract (str):
                The method used for extracting the line data. See the docstring
                of the grid method `get_line_data` to find supported values.
        
        Returns:
            dict: Information useful for performing a line plot of the field
        """
        # turn field into scalar field
        scalar_data = self.to_scalar(scalar).data
        
        # extract the line data
        data = self.grid.get_line_data(scalar_data, extract=extract)
        if 'label_y' in data and data['label_y']:
            if self.label:
                data['label_y'] = f"{self.label} ({data['label_y']})"
        else:
            data['label_y'] = self.label
        return data  # type: ignore  
        
                
    def get_image_data(self, scalar: str = 'auto',
                       transpose: bool = False,
                       **kwargs) -> Dict[str, Any]:
        r""" return data for plotting an image of the field
        
        Args:
            scalar (str or int):
                The method for extracting scalars as described in
                :meth:`DataFieldBase.to_scalar`.
            transpose (bool):
                Determines whether the transpose of the data should is plotted
            \**kwargs:
                Additional parameters are forwarded to `grid.get_image_data`
        
        Returns:
            dict: Information useful for plotting an image of the field        
        """
        # turn field into scalar field
        scalar_data = self.to_scalar(scalar).data
        
        # extract the image data
        data = self.grid.get_image_data(scalar_data, **kwargs)  # type: ignore 
        data['title'] = self.label
        
        if transpose:
            # adjust image data such that the transpose is plotted
            data['data'] = data['data'].T
            data['label_x'], data['label_y'] = data['label_y'], data['label_x']
            
        return data  # type: ignore  

    
    def get_vector_data(self, **kwargs) -> Dict[str, Any]:
        r""" return data for a vector plot of the field
        
        Args:
            \**kwargs: Additional parameters are forwarded to
                `grid.get_image_data`
        
        Returns:
            dict: Information useful for plotting an vector field        
        """
        raise NotImplementedError()
    
                
    def _plot_line(self, ax,
                   extract: str = 'auto',
                   ylabel: str = None,
                   **kwargs) -> PlotReference:
        r""" visualize a field using a 1d line plot
        
        Args:
            ax (:class:`matplotlib.axes.Axes`):
                Figure axes to be used for plotting.
            extract (str):
                The method used for extracting the line data.
            ylabel (str):
                Label of the y-axis. If omitted, the label is chosen 
                automatically from the data field.
            \**kwargs:
                Additional keyword arguments are passed to
                :func:`matplotlib.pyplot.plot`
                
        Returns:
            :class:`PlotReference`: Instance that contains information to update
            the plot with new data later.
        """
        # obtain data for the plot
        line_data = self.get_line_data(extract=extract)
        
        # do the plot
        line2d, = ax.plot(line_data['data_x'], line_data['data_y'],
                              **kwargs)

        # set some default properties
        ax.set_xlabel(line_data['label_x'])
        if ylabel is None:
            ylabel = line_data.get('label_y', self.label)
        if ylabel:
            ax.set_ylabel(ylabel)
                
        return PlotReference(ax, line2d, {'extract': extract})
    
    
    def _update_line_plot(self, reference: PlotReference) -> None:
        """ update a line plot with the current field values
        
        Args:
            reference (:class:`PlotReference`):
                The reference to the plot that is updated
        """
        import matplotlib as mpl
        
        # obtain data for the plot
        extract = reference.parameters.get('extract', 'auto')
        line_data = self.get_line_data(extract=extract)
        
        line2d = reference.element
        if isinstance(line2d, mpl.lines.Line2D):
            # update old plot
            line2d.set_xdata(line_data['data_x'])
            line2d.set_ydata(line_data['data_y'])
            
        else:
            raise ValueError(f'Unsupported plot reference {reference}')


    def _plot_image(self, ax,
                    colorbar: bool = False,
                    scalar: str = 'auto',
                    transpose: bool = False,
                    **kwargs) -> PlotReference:
        r""" visualize a field using a 2d density plot

        Args:
            ax (:class:`matplotlib.axes.Axes`):
                Figure axes to be used for plotting.
            colorbar (bool):
                Determines whether a colorbar is shown
            scalar (str or int):
                The method for extracting scalars as described in
                :meth:`DataFieldBase.to_scalar`.
            transpose (bool):
                Determines whether the transpose of the data should is plotted
            \**kwargs:
                Additional keyword arguments that affect the image. For
                instance, some fields support a `scalar` argument that
                determines how they are converted to a scalar. Non-Cartesian
                grids might support a `performance_goal` argument to influence
                how an image is created from the raw data. Finally, the
                remaining arguments are are passed to
                :func:`matplotlib.pyplot.imshow` to affect the appearance.
                
        Returns:
            :class:`PlotReference`: Instance that contains information to update
            the plot with new data later.
        """
        # obtain image data
        data = self.get_image_data(scalar, transpose)
        
        if ax is None:
            import matplotlib.pyplot as plt
            # create new figure
            ax = plt.subplots()[1]

        kwargs.setdefault('origin', 'lower')
        kwargs.setdefault('interpolation', 'none')
        axes_image = ax.imshow(data['data'], extent=data['extent'], **kwargs)
        
        # set some default properties
        ax.set_xlabel(data['label_x'])
        ax.set_ylabel(data['label_y'])
        ax.set_title(data.get('title', self.label))
        
        if colorbar:
            from ..tools.misc import add_scaled_colorbar
            add_scaled_colorbar(axes_image, ax=ax)
                
        parameters = {'scalar': scalar, 'transpose': transpose}
        return PlotReference(ax, axes_image, parameters)

    
    def _update_image_plot(self, reference: PlotReference) -> None:
        """ update an image plot with the current field values
        
        Args:
            reference (:class:`PlotReference`):
                The reference to the plot that is updated
        """
        # obtain image data
        p = reference.parameters
        data = self.get_image_data(scalar=p.get('scalar', 'auto'),
                                   transpose=p.get('transpose', False))
        
        # update the axes image
        reference.element.set_array(data['data'])


    def _plot_vector(self, ax,
                     method: str = 'quiver',
                     transpose: bool = False,
                     max_points: int = 16,
                     **kwargs) -> PlotReference:
        r""" visualize a field using a 2d vector plot

        Args:
            ax (:class:`matplotlib.axes.Axes`):
                Figure axes to be used for plotting.
            method (str):
                Plot type that is used. This can be either `quiver` or
                `streamplot`.
            transpose (bool):
                Determines whether the transpose of the data should be plotted.
            max_points (int):
                The maximal number of points that is used along each axis. This
                argument is only used for quiver plots.
            \**kwargs:
                Additional keyword arguments are passed to
                :func:`matplotlib.pyplot.quiver` or
                :func:`matplotlib.pyplot.streamplot`.
                
        Returns:
            :class:`PlotReference`: Instance that contains information to update
            the plot with new data later.
        """
        # do the plotting using the chosen method
        if method == 'quiver':
            data = self.get_vector_data(transpose=transpose,
                                        max_points=max_points)
            element = ax.quiver(data['x'], data['y'], data['data_x'],
                                data['data_y'], **kwargs)        
            
        elif method == 'streamplot':
            data = self.get_vector_data(transpose=transpose)
            element = ax.streamplot(data['x'], data['y'], data['data_x'],
                                    data['data_y'], **kwargs)        
                
        else:
            raise ValueError(f'Vector plot `{method}` is not supported.')
            
        # set some default properties of the plot
        ax.set_aspect('equal')
        ax.set_xlabel(data['label_x'])
        ax.set_ylabel(data['label_y'])
        ax.set_title(data.get('title', self.label))

        parameters = {'method': method, 'transpose': transpose,
                      'max_points': max_points, 'kwargs': kwargs}
        return PlotReference(ax, element, parameters)

    
    def _update_vector_plot(self, reference: PlotReference) -> None:
        """ update a vector plot with the current field values
        
        Args:
            reference (:class:`PlotReference`):
                The reference to the plot that is updated
        """
        method = reference.parameters.get('method', 'quiver')
        transpose = reference.parameters.get('transpose', False)
        if method == 'quiver':
            max_points = reference.parameters.get('max_points')
            data = self.get_vector_data(transpose=transpose,
                                        max_points=max_points)
            reference.element.set_UVC(data['data_x'], data['data_y'])
            
        elif method == 'streamplot':
            ax = reference.ax
            kwargs = reference.parameters.get('kwargs', {})
            data = self.get_vector_data(transpose=transpose)
            # remove old streamplot
            reference.element.remove()  
            # update with new streamplot
            reference.element = ax.streamplot(data['x'], data['y'],
                                              data['data_x'], data['data_y'],
                                              **kwargs)        
                
        else:
            raise ValueError(f'Vector plot `{method}` is not supported.')

    
    def _update_plot(self, reference: PlotReference) -> None:
        """ update a plot with the current field values
        
        Args:
            reference (:class:`PlotReference`):
                The reference to the plot to updated
        """
        import matplotlib as mpl
        
        # update the plot based on the given reference
        el = reference.element
        if isinstance(el, mpl.lines.Line2D):
            self._update_line_plot(reference)
        elif isinstance(el, mpl.image.AxesImage):
            self._update_image_plot(reference)
        elif isinstance(el, (mpl.quiver.Quiver, mpl.streamplot.StreamplotSet)):
            self._update_vector_plot(reference)
        else:
            raise ValueError(f'Unknown plot element {el.__class__.__name__}')
    

    @plot_on_axes(update_method='_update_plot')
    def plot(self, ax, kind: str = 'auto', **kwargs):
        r""" visualize the field
        
        Args:
            ax (:class:`matplotlib.axes.Axes`):
                Figure axes to be used for plotting. If `None`, a new figure is
                created. This has no effect if a `reference` is supplied.
            kind (str):
                Determines the visualizations. Supported values are `image`, 
                `line`, `vector`, or `interactive`. Alternatively, `auto`
                determines the best visualization based on the field itself.
            \**kwargs:
                All additional keyword arguments are forwarded to the actual
                plotting function.
                
        Returns:
            :class:`PlotReference`: Instance that contains information to update
            the plot with new data later.
        """
        # pre-process the kinds
        if kind == 'auto':
            # determine best plot for this field
            if (isinstance(self, DataFieldBase) and self.rank == 1 and
                    self.grid.dim == 2):
                kind = 'vector'
            elif len(self.grid.shape) == 1:
                kind = 'line'
            else:
                kind = 'image'
                
        elif kind == 'quiver':
            kind = 'vector'
            kwargs['method'] = 'quiver'

        elif kind == 'streamplot':
            kind = 'vector'
            kwargs['method'] = 'streamplot'

        # do the actual plotting
        if kind == 'image':
            reference = self._plot_image(ax=ax, **kwargs)
        elif kind == 'line':
            reference = self._plot_line(ax=ax, **kwargs)
        elif kind == 'vector':
            reference = self._plot_vector(ax=ax, **kwargs)
        else:
            raise ValueError(f'Unsupported plot `{kind}`. Possible choices '
                             'are `image`, `line`, `vector`, or `auto`.')

        return reference

    
    def plot_interactive(self, scalar: str = 'auto', **kwargs):
        """ create an interactive plot of the field using :mod:`napari`
        
        Args:
            scalar (str): The method for obtaining scalar values of fields        
            **kwargs: Extra arguments are passed to :class:`napari.Viewer`
        """
        from ..tools.plotting import napari_viewer
        
        with napari_viewer(self.grid, **kwargs) as viewer:
            viewer.add_image(self.to_scalar(scalar).data,
                             name=self.label,
                             rgb=False,
                             scale=self.grid.discretization)
    


'''
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
'''

import pytest
import numpy as np

from ... import pdes
from ...grids import UnitGrid
from ...fields import ScalarField



@pytest.mark.parametrize('dim', [1, 2])
@pytest.mark.parametrize('pde_class', [pdes.KuramotoSivashinskyPDE, 
                                       pdes.KPZInterfacePDE,
                                       pdes.SwiftHohenbergPDE,
                                       pdes.DiffusionPDE,
                                       pdes.CahnHilliardPDE])
def test_pde_consistency(pde_class, dim):
    """ test some methods of generic PDE models """
    eq = pde_class()
    assert isinstance(str(eq), str)
    assert isinstance(repr(eq), str)
    
    grid = UnitGrid([4] * dim)
    state = ScalarField.random_uniform(grid)
    field = eq.evolution_rate(state)
    assert field.grid == grid
    rhs = eq._make_pde_rhs_numba(state)
    np.testing.assert_allclose(field.data, rhs(state.data, 0))
    


def test_pde_consistency_test():
    """ test whether the consistency of a pde implementation is checked """

    class TestPDE(pdes.PDEBase):
        def evolution_rate(self, field, t=0):
            return 2 * field
        
        def _make_pde_rhs_numba(self, state):
            def impl(state_data, t):
                return 3 * state_data
            return impl
        
    eq = TestPDE()
    state = ScalarField.random_uniform(UnitGrid([4]))
    with pytest.raises(RuntimeError):
        eq.solve(state, t_range=5)
    


#!/usr/bin/env python3

import sys
from collections import OrderedDict
from pathlib import Path

PACKAGE_PATH = Path(__file__).resolve().parents[1]
sys.path.append(str(PACKAGE_PATH))

from pde import environment

env = environment(OrderedDict)

for category, data in env.items():
    if hasattr(data, 'items'):
        print(f'\n{category}:')
        for key, value in data.items():
            print(f'    {key}: {value}')
    else:
        data_formatted = data.replace("\n", "\n    ")
        print(f'{category}: {data_formatted}')



"""
Diffusion on a Cartesian grid
=============================

This example shows how to solve the diffusion equation on a Cartesian grid.
"""

from pde import CartesianGrid, ScalarField, DiffusionPDE

grid = CartesianGrid([[-1, 1], [0, 2]], [15, 8])  # generate grid
state = ScalarField(grid)  # generate initial condition
state.add_interpolated([0, 1], 1)

eq = DiffusionPDE()  # define the pde
result = eq.solve(state, t_range=1, dt=0.005)
result.plot(show=True)



r"""
Plotting a vector field
=======================

This example shows how to initialize and visualize the vector field 
:math:`\boldsymbol u = \bigl(\sin(x), \cos(x)\bigr)`.
"""

from pde import CartesianGrid, VectorField

grid = CartesianGrid([[-2, 2], [-2, 2]], 32)
field = VectorField.from_expression(grid, ['sin(x)', 'cos(x)'])
field.plot(method='streamplot', title='Stream plot', show=True)
